#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from functools import partial
from typing import Optional

from merlin.models.tf.blocks.interaction import DotProductInteraction
from merlin.models.tf.core.aggregation import StackFeatures
from merlin.models.tf.core.base import Block, Debug
from merlin.models.tf.core.combinators import Filter, ParallelBlock, SequentialBlock
from merlin.models.tf.inputs.continuous import ContinuousFeatures
from merlin.models.tf.inputs.embedding import EmbeddingOptions, Embeddings
from merlin.models.utils.schema_utils import infer_embedding_dim
from merlin.schema import Schema, Tags


def DLRMBlock(
    schema: Schema,
    *,
    embedding_dim: int = None,
    embedding_options: EmbeddingOptions = None,
    bottom_block: Optional[Block] = None,
    top_block: Optional[Block] = None,
    embeddings: Optional[Block] = None,
) -> SequentialBlock:
    """Builds the DLRM architecture, as proposed in the following
    `paper https://arxiv.org/pdf/1906.00091.pdf`_ [1]_.

    References
    ----------
    .. [1] Naumov, Maxim, et al. "Deep learning recommendation model for
       personalization and recommendation systems." arXiv preprint arXiv:1906.00091 (2019).

    Parameters
    ----------
    schema : Schema
        The `Schema` with the input features
    bottom_block : Block
        The `Block` that combines the continuous features (typically a `MLPBlock`)
    top_block : Optional[Block], optional
        The optional `Block` that combines the outputs of bottom layer and of
        the factorization machine layer, by default None
    embedding_dim : Optional[int], optional
        Dimension of the embeddings, by default None
    embedding_options : EmbeddingOptions
        Options for the input embeddings.
        - embedding_dim_default: int - Default dimension of the embedding
        table, when the feature is not found in ``embedding_dims``, by default 64
        - infer_embedding_sizes : bool, Automatically defines the embedding
        dimension from the feature cardinality in the schema, by default False,
        which needs to be kept False for the DLRM architecture.

    Returns
    -------
    SequentialBlock
        The DLRM block

    Raises
    ------
    ValueError
        The schema is required by DLRM
    ValueError
        The bottom_block is required by DLRM
    ValueError
        The embedding_dim (X) needs to match the last layer of bottom MLP (Y).
    ValueError
        Only one-of `embeddings` or `embedding_options` can be used.
    """
    if schema is None:
        raise ValueError("The schema is required by DLRM")

    con_schema = schema.select_by_tag(Tags.CONTINUOUS).excluding_by_tag(Tags.TARGET)
    cat_schema = schema.select_by_tag(Tags.CATEGORICAL).excluding_by_tag(Tags.TARGET)

    if not len(cat_schema) > 0:
        raise ValueError("DLRM requires categorical features")

    if embeddings is not None and embedding_options is not None:
        raise ValueError("Only one-of `embeddings` or `embedding_options` may be provided.")

    if embeddings is None:
        embeddings = _get_embeddings(embedding_dim, embedding_options, bottom_block, cat_schema)

    if len(con_schema) > 0:
        if bottom_block is None:
            raise ValueError(
                "The bottom_block is required by DLRM when "
                "continuous features are available in the schema"
            )
        con = ContinuousFeatures.from_schema(con_schema)
        bottom_block = con.connect(bottom_block)  # type: ignore
        interaction_inputs = ParallelBlock({"embeddings": embeddings, "bottom_block": bottom_block})
    else:
        interaction_inputs = embeddings  # type: ignore
        bottom_block = None

    interaction_inputs = interaction_inputs.connect(Debug())

    if not top_block:
        return interaction_inputs.connect(DotProductInteractionBlock())

    if not bottom_block:
        return interaction_inputs.connect(DotProductInteractionBlock(), top_block)

    top_block_inputs = interaction_inputs.connect_with_shortcut(
        DotProductInteractionBlock(),
        shortcut_filter=Filter("bottom_block"),
        aggregation="concat",
    )
    top_block_outputs = top_block_inputs.connect(top_block)

    return top_block_outputs


def _get_embeddings(embedding_dim, embedding_options, bottom_block, cat_schema):
    if embedding_dim is None:
        raise ValueError("The embedding_dim is required")

    if embedding_options is not None:
        embedding_options.embedding_dim_default = embedding_dim
    else:
        embedding_options = EmbeddingOptions(embedding_dim_default=embedding_dim)

    if (
        embedding_dim is not None
        and bottom_block is not None
        and embedding_dim != bottom_block.layers[-1].units
    ):
        raise ValueError(
            f"The embedding_dim ({embedding_dim}) needs to match the "
            "last layer of bottom MLP ({bottom_block.layers[-1].units}) "
        )

    embeddings_kwargs = dict(
        sequence_combiner=embedding_options.combiner,
        embeddings_initializer=embedding_options.embeddings_initializers,
        dim=embedding_options.embedding_dim_default,
        infer_dim_fn=partial(
            infer_embedding_dim,
            multiplier=embedding_options.infer_embedding_sizes_multiplier,
            ensure_multiple_of_8=embedding_options.infer_embeddings_ensure_dim_multiple_of_8,
        ),
    )

    return Embeddings(cat_schema, **embeddings_kwargs)


def DotProductInteractionBlock():
    return SequentialBlock(StackFeatures(axis=1), DotProductInteraction())



from testbook import testbook

from tests.conftest import REPO_ROOT


@testbook(
    REPO_ROOT / "examples/usecases/entertainment-with-pretrained-embeddings.ipynb", execute=False
)
def test_usecase_pretrained_embeddings(tb):
    tb.execute()
    model = tb.ref("model")
    assert set(model.history.history.keys()) == set(["auc", "loss", "regularization_loss"])



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# import pytest
# import merlin.models.torch as ml

# examples_utils = pytest.importorskip("merlin.models.torch.losses")


# @pytest.mark.parametrize("label_smoothing", [0.0, 0.1, 0.6])
# def test_item_prediction_with_label_smoothing_ce_loss(
#     torch_yoochoose_tabular_transformer_features, torch_yoochoose_like, label_smoothing
# ):
#     np = pytest.importorskip("numpy")
#     custom_loss = examples_utils.LabelSmoothCrossEntropyLoss(
#         reduction="mean", smoothing=label_smoothing
#     )
#     input_module = torch_yoochoose_tabular_transformer_features
#     body = ml.SequentialBlock(input_module, ml.MLPBlock([64]))
#     head = ml.Head(
#         body, ml.NextItemPredictionTask(weight_tying=True, loss=custom_loss), inputs=input_module
#     )
#
#     body_outputs = body(torch_yoochoose_like)
#
#     trg_flat = input_module.masking.masked_targets.flatten()
#     non_pad_mask = trg_flat != input_module.masking.padding_idx
#     labels_all = torch.masked_select(trg_flat, non_pad_mask)
#     predictions = head(body_outputs)
#
#     loss = head.prediction_task_dict["next-item"].compute_loss(
#         inputs=body_outputs,
#         targets=labels_all,
#     )
#
#     n_classes = 51997
#     manuall_loss = torch.nn.NLLLoss(reduction="mean")
#     target_with_smoothing = labels_all * (1 - label_smoothing) + label_smoothing / n_classes
#     manual_output_loss = manuall_loss(predictions, target_with_smoothing.to(torch.long))
#
#     assert np.allclose(manual_output_loss.detach().numpy(), loss.detach().numpy(), rtol=1e-3)



# Test is currently breaks in TF 2.10

from testbook import testbook

from tests.conftest import REPO_ROOT

p = "examples/usecases/ecommerce-session-based-next-item-prediction-for-fashion.ipynb"


@testbook(
    REPO_ROOT / p,
    timeout=180,
    execute=False,
)
def test_usecase_ecommerce_session_based(tb):
    tb.inject(
        """
        import os
        from unittest.mock import patch
        from merlin.datasets.synthetic import generate_data
        mock_train, mock_valid = generate_data(
            input="dressipi2022-preprocessed",
            num_rows=10000,
            set_sizes=(0.8, 0.2)
        )
        p1 = patch(
            "merlin.datasets.ecommerce.get_dressipi2022",
            return_value=[mock_train, mock_valid]
        )
        p1.start()
        os.environ["DATA_FOLDER"] = "/tmp/dressipi2022/"
        os.environ["EPOCHS"] = "1"
        """
    )
    tb.execute()
    metrics_mlp = tb.ref("metrics_mlp")
    assert set(metrics_mlp.keys()) == set(
        [
            "loss",
            "recall_at_100",
            "mrr_at_100",
            "ndcg_at_100",
            "map_at_100",
            "precision_at_100",
            "regularization_loss",
        ]
    )
    metrics_bi_lstm = tb.ref("metrics_bi_lstm")
    assert set(metrics_bi_lstm.keys()) == set(
        [
            "loss",
            "recall_at_100",
            "mrr_at_100",
            "ndcg_at_100",
            "map_at_100",
            "precision_at_100",
            "regularization_loss",
        ]
    )



import os

from testbook import testbook

from tests.conftest import REPO_ROOT


@testbook(REPO_ROOT / "examples/03-Exploring-different-models.ipynb", execute=False)
def test_example_03_exploring_different_models(tb):
    tb.inject(
        """
        import os
        os.environ["DATA_FOLDER"] = "/tmp/data/"
        os.environ["NUM_ROWS"] = "999"
        """
    )
    NUM_OF_CELLS = len(tb.cells)
    tb.execute_cell(list(range(0, NUM_OF_CELLS - 5)))
    metrics_ncf = tb.ref("metrics_ncf")
    assert set(metrics_ncf.keys()) == set(
        [
            "auc",
            "loss",
            "regularization_loss",
        ]
    )
    metrics_mlp = tb.ref("metrics_mlp")
    assert set(metrics_mlp.keys()) == set(
        [
            "auc_1",
            "loss",
            "regularization_loss",
        ]
    )
    metrics_wide_n_deep = tb.ref("metrics_wide_n_deep")
    assert set(metrics_wide_n_deep.keys()) == set(
        [
            "auc_2",
            "loss",
            "regularization_loss",
        ]
    )
    metrics_dlrm = tb.ref("metrics_dlrm")
    assert set(metrics_dlrm.keys()) == set(
        [
            "auc_3",
            "loss",
            "regularization_loss",
        ]
    )
    metrics_dcn = tb.ref("metrics_dcn")
    assert set(metrics_dcn.keys()) == set(
        [
            "auc_4",
            "loss",
            "regularization_loss",
        ]
    )
    assert os.path.isfile("results.txt")
    tb.execute_cell(NUM_OF_CELLS - 2)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import typing
from typing import Dict

import tensorflow as tf

# TODO: Make this more generic and work with multi-hot features
TabularData = Dict[str, tf.Tensor]
TensorOrTabularData = typing.Union[tf.Tensor, TabularData]
LossReduction = typing.Callable[[typing.List[tf.Tensor]], tf.Tensor]

__all__ = [
    "TabularData",
    "TensorOrTabularData",
    "LossReduction",
]



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest
import tensorflow as tf

import merlin.models.tf as mm
from merlin.io import Dataset
from merlin.models.tf.outputs.classification import CategoricalTarget, EmbeddingTablePrediction
from merlin.models.tf.utils import testing_utils
from merlin.schema import Tags


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_binary_output(ecommerce_data: Dataset, run_eagerly):
    model = mm.Model(
        mm.InputBlock(ecommerce_data.schema),
        mm.MLPBlock([8]),
        mm.BinaryOutput("click"),
    )

    _, history = testing_utils.model_test(model, ecommerce_data, run_eagerly=run_eagerly)

    assert set(history.history.keys()) == {
        "loss",
        "precision",
        "recall",
        "binary_accuracy",
        "auc",
        "regularization_loss",
    }


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_categorical_output(sequence_testing_data: Dataset, run_eagerly):
    dataloader, schema = _next_item_loader(sequence_testing_data)
    model = mm.Model(
        mm.InputBlock(schema),
        mm.MLPBlock([8]),
        mm.CategoricalOutput(schema["item_id_seq"]),
    )

    _, history = testing_utils.model_test(model, dataloader, run_eagerly=run_eagerly)

    assert set(history.history.keys()) == {
        "loss",
        "ndcg_at_10",
        "precision_at_10",
        "map_at_10",
        "mrr_at_10",
        "recall_at_10",
        "regularization_loss",
    }


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_next_item_prediction(sequence_testing_data: Dataset, run_eagerly):
    dataloader, schema = _next_item_loader(sequence_testing_data)
    embeddings = mm.Embeddings(
        schema,
        sequence_combiner=tf.keras.layers.Lambda(lambda x: tf.reduce_mean(x, axis=1)),
    )

    predictions = [
        schema.select_by_name("item_id_seq"),
        schema["item_id_seq"],
        CategoricalTarget(schema["item_id_seq"]),
        embeddings["item_id_seq"],
        EmbeddingTablePrediction(embeddings["item_id_seq"]),
    ]

    for target in predictions:
        model = mm.Model(
            mm.InputBlockV2(schema, categorical=embeddings),
            mm.MLPBlock([32]),
            mm.CategoricalOutput(target),
        )
        testing_utils.model_test(model, dataloader, run_eagerly=run_eagerly)


def _next_item_loader(sequence_testing_data: Dataset):
    def _last_interaction_as_target(inputs, targets):
        inputs = mm.ListToRagged()(inputs)
        items = inputs["item_id_seq"]
        _items = items[:, :-1]
        targets = tf.one_hot(items[:, -1:].flat_values, 51997)
        inputs["item_id_seq"] = _items
        return inputs, targets

    schema = sequence_testing_data.schema.select_by_tag(Tags.CATEGORICAL)
    sequence_testing_data.schema = schema
    dataloader = mm.Loader(sequence_testing_data, batch_size=50)
    dataloader = dataloader.map(_last_interaction_as_target)
    return dataloader, schema



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import warnings
from typing import Union

import numpy as np
import pandas as pd
from scipy.sparse import coo_matrix

from merlin.io import Dataset
from merlin.schema import Tags


def dataset_to_coo(dataset: Dataset):
    """Converts a merlin.io.Dataset object to a scipy coo matrix"""
    user_id_column = dataset.schema.select_by_tag(Tags.USER_ID).first.name
    item_id_column = dataset.schema.select_by_tag(Tags.ITEM_ID).first.name

    columns = [user_id_column, item_id_column]
    target_column = None
    target = dataset.schema.select_by_tag(Tags.TARGET)

    if len(target) > 1:
        raise ValueError(
            "Found more than one column tagged Tags.TARGET in the dataset schema."
            f" Expected a single target column but found  {target.column_names}"
        )

    elif len(target) == 1:
        target_column = target.first.name
        columns.append(target_column)

    df = dataset.to_ddf()[columns].compute(scheduler="synchronous")

    userids = _to_numpy(df[user_id_column])
    itemids = _to_numpy(df[item_id_column])
    targets = _to_numpy(df[target_column]) if target_column else np.ones(len(userids))
    return coo_matrix((targets.astype("float32"), (userids, itemids)))


def unique_rows_by_features(
    dataset: Dataset, features_tag: Union[str, Tags], grouping_tag: Union[str, Tags] = Tags.ID
) -> Dataset:
    """
    Select unique rows from a Dataset. Returns columns specified by `features_tag`
     that are unique based on the columns specified by the `grouping_tag`.

    Parameters
    ----------
    dataset : ~merlin.io.Dataset
        Dataset to transform
    features_tag : ~merlin.schema.Tags
        Tag representing the columns to return in the new Dataset
    grouping_tag : ~merlin.schema.Tags
        Tag representing the columns to check for uniqueness.
        Default: Tags.ID

    Returns
    -------
        Dataset
    """
    warnings.warn(
        "`unique_rows_by_features` is deprecated and will be removed in a future version. "
        "Please use `unique_by_tag` instead.",
        DeprecationWarning,
    )
    return unique_by_tag(dataset, features_tag, grouping_tag)


def unique_by_tag(
    dataset: Dataset, features_tag: Union[str, Tags], grouping_tag: Union[str, Tags] = Tags.ID
) -> Dataset:
    """
    Select unique rows from a Dataset. Returns columns specified by `features_tag`
     that are unique based on the columns specified by the `grouping_tag`.

    Parameters
    ----------
    dataset : ~merlin.io.Dataset
        Dataset to transform
    features_tag : ~merlin.schema.Tags
        Tag representing the columns to return in the new Dataset
    grouping_tag : ~merlin.schema.Tags
        Tag representing the columns to check for uniqueness.
        Default: Tags.ID

    Returns
    -------
        Dataset
    """
    ddf = dataset.to_ddf()

    features_schema = dataset.schema.select_by_tag(features_tag)
    columns = features_schema.column_names

    if columns:
        id_col = features_schema.select_by_tag(grouping_tag).first.name
        ddf = ddf[columns].drop_duplicates(id_col, keep="first")

    return Dataset(ddf, schema=features_schema)


def _to_numpy(series):
    """converts a pandas or cudf series to a numpy array"""
    if isinstance(series, pd.Series):
        return series.values
    else:
        return series.values_host



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Callable, Optional

import fiddle as fdl

from merlin.io import Dataset
from tests.common.tf.retrieval.retrieval_config import config_retrieval_train_eval_runner
from tests.common.tf.tests_utils import extract_hparams_from_config


def set_lastfm_two_tower_hparams_config(runner_cfg: fdl.Config):
    runner_cfg.model.two_tower_activation = "selu"
    runner_cfg.model.two_tower_mlp_layers = "64"
    runner_cfg.model.two_tower_dropout = 0.2
    runner_cfg.model.two_tower_embedding_sizes_multiplier = 4.0
    runner_cfg.model.logits_temperature = 0.6
    runner_cfg.model.l2_reg = 2e-06
    runner_cfg.model.embeddings_l2_reg = 1e-05
    runner_cfg.model.logq_correction_factor = 0.7

    runner_cfg.model.samplers.neg_sampling = "inbatch"

    runner_cfg.loss.loss = "categorical_crossentropy"
    runner_cfg.loss.xe_label_smoothing = 0.3

    runner_cfg.optimizer.lr = 0.02
    runner_cfg.optimizer.lr_decay_rate = 0.97
    runner_cfg.optimizer.lr_decay_steps = 50
    runner_cfg.optimizer.optimizer = "adam"

    runner_cfg.metrics.topk_metrics_cutoffs = "10,50,100"

    runner_cfg.train_batch_size = 4096
    runner_cfg.eval_batch_size = 512

    runner_cfg.train_epochs = 20
    runner_cfg.train_steps_per_epoch = None
    runner_cfg.eval_steps = 5000

    runner_cfg.callbacks.train_batch_size = runner_cfg.train_batch_size


def set_lastfm_mf_hparams_config(runner_cfg: fdl.Config):
    runner_cfg.model.mf_dim = 64
    runner_cfg.model.logits_temperature = 1.4
    runner_cfg.model.embeddings_l2_reg = 3e-07
    runner_cfg.model.logq_correction_factor = 0.9

    runner_cfg.model.samplers.neg_sampling = "inbatch"

    runner_cfg.loss.loss = "categorical_crossentropy"
    runner_cfg.loss.xe_label_smoothing = 0.0

    runner_cfg.optimizer.lr = 0.005
    runner_cfg.optimizer.lr_decay_rate = 0.98
    runner_cfg.optimizer.lr_decay_steps = 50
    runner_cfg.optimizer.optimizer = "adam"

    runner_cfg.metrics.topk_metrics_cutoffs = "10,50,100"

    runner_cfg.train_batch_size = 4096
    runner_cfg.eval_batch_size = 512

    runner_cfg.train_epochs = 20
    runner_cfg.train_steps_per_epoch = None
    runner_cfg.eval_steps = 5000


def train_eval_two_tower(
    train_ds: Dataset,
    eval_ds: Dataset,
    train_epochs: int = 1,
    train_steps_per_epoch: Optional[int] = None,
    eval_steps: Optional[int] = 2000,
    train_batch_size: int = 512,
    eval_batch_size: int = 512,
    topk_metrics_cutoffs: str = "10,50,100",
    log_to_wandb: bool = False,
    wandb_project: str = None,
    config_callback: Callable = None,
):
    runner_cfg = config_retrieval_train_eval_runner(
        train_ds,
        eval_ds,
        model_type="two_tower",
        log_to_wandb=log_to_wandb,
        wandb_project=wandb_project,
    )

    if config_callback:
        config_callback(runner_cfg)

    runner_cfg.train_epochs = train_epochs
    runner_cfg.train_steps_per_epoch = train_steps_per_epoch
    runner_cfg.eval_steps = eval_steps
    runner_cfg.train_batch_size = train_batch_size
    runner_cfg.eval_batch_size = eval_batch_size
    runner_cfg.metrics.topk_metrics_cutoffs = topk_metrics_cutoffs

    hparams = extract_hparams_from_config(runner_cfg)

    runner = fdl.build(runner_cfg)
    metrics = runner.run(hparams)
    return metrics


def train_eval_mf(
    train_ds: Dataset,
    eval_ds: Dataset,
    train_epochs: int = 1,
    train_steps_per_epoch: Optional[int] = None,
    eval_steps: Optional[int] = 2000,
    train_batch_size: int = 512,
    eval_batch_size: int = 512,
    topk_metrics_cutoffs: str = "10,50,100",
    log_to_wandb: bool = False,
    wandb_project: str = None,
    config_callback: Callable = None,
):
    runner_cfg = config_retrieval_train_eval_runner(
        train_ds,
        eval_ds,
        model_type="mf",
        log_to_wandb=log_to_wandb,
        wandb_project=wandb_project,
    )

    if config_callback:
        config_callback(runner_cfg)

    runner_cfg.train_epochs = train_epochs
    runner_cfg.train_steps_per_epoch = train_steps_per_epoch
    runner_cfg.eval_steps = eval_steps
    runner_cfg.train_batch_size = train_batch_size
    runner_cfg.eval_batch_size = eval_batch_size
    runner_cfg.metrics.topk_metrics_cutoffs = topk_metrics_cutoffs

    runner_cfg.callbacks.train_batch_size = runner_cfg.train_batch_size

    hparams = extract_hparams_from_config(runner_cfg)

    runner = fdl.build(runner_cfg)
    metrics = runner.run(hparams)
    return metrics


def train_eval_two_tower_for_lastfm(
    train_ds: Dataset,
    eval_ds: Dataset,
    train_epochs: int = 1,
    train_steps_per_epoch: Optional[int] = None,
    eval_steps: Optional[int] = 2000,
    train_batch_size: int = 512,
    eval_batch_size: int = 512,
    topk_metrics_cutoffs: str = "10,50,100",
    log_to_wandb: bool = False,
    wandb_project: str = None,
):
    return train_eval_two_tower(
        train_ds,
        eval_ds,
        train_epochs,
        train_steps_per_epoch,
        eval_steps,
        train_batch_size,
        eval_batch_size,
        topk_metrics_cutoffs,
        log_to_wandb,
        wandb_project,
        config_callback=set_lastfm_two_tower_hparams_config,
    )


def train_eval_mf_for_lastfm(
    train_ds: Dataset,
    eval_ds: Dataset,
    train_epochs: int = 1,
    train_steps_per_epoch: Optional[int] = None,
    eval_steps: Optional[int] = 2000,
    train_batch_size: int = 512,
    eval_batch_size: int = 512,
    topk_metrics_cutoffs: str = "10,50,100",
    log_to_wandb: bool = False,
    wandb_project: str = None,
):
    return train_eval_mf(
        train_ds,
        eval_ds,
        train_epochs,
        train_steps_per_epoch,
        eval_steps,
        train_batch_size,
        eval_batch_size,
        topk_metrics_cutoffs,
        log_to_wandb,
        wandb_project,
        config_callback=set_lastfm_mf_hparams_config,
    )



import numpy as np
import pytest
import tensorflow as tf
from transformers import BertConfig

import merlin.models.tf as mm
from merlin.io import Dataset
from merlin.models.tf.loader import Loader
from merlin.models.tf.transformers.block import (
    AlbertBlock,
    BertBlock,
    GPT2Block,
    RobertaBlock,
    XLNetBlock,
)
from merlin.models.tf.utils import testing_utils
from merlin.schema import Tags


def test_import():
    import transformers

    assert transformers is not None


def test_transformer_encoder():
    NUM_ROWS = 100
    SEQ_LENGTH = 10
    EMBED_DIM = 128

    inputs = tf.random.uniform((NUM_ROWS, SEQ_LENGTH, EMBED_DIM))
    transformer_encod = mm.TransformerBlock(
        transformer=BertConfig(hidden_size=EMBED_DIM, num_attention_heads=16)
    )
    outputs = transformer_encod(inputs)

    assert list(outputs.shape) == [NUM_ROWS, SEQ_LENGTH, EMBED_DIM]


def test_transformer_encoder_with_pooling():
    NUM_ROWS = 100
    SEQ_LENGTH = 10
    EMBED_DIM = 128

    inputs = tf.random.uniform((NUM_ROWS, SEQ_LENGTH, EMBED_DIM))
    transformer_encod = mm.TransformerBlock(
        transformer=BertConfig(hidden_size=EMBED_DIM, num_attention_heads=16),
        transformer_post="pooler_output",
    )
    outputs = transformer_encod(inputs)

    assert list(outputs.shape) == [NUM_ROWS, EMBED_DIM]


@pytest.mark.parametrize("max_seq_length", [None, 5, 20])
def test_transformer_encoder_with_list_to_dense(max_seq_length):
    NUM_ROWS = 100
    SEQ_LENGTH = 10
    EMBED_DIM = 128
    inputs = tf.RaggedTensor.from_tensor(tf.random.uniform((NUM_ROWS, SEQ_LENGTH, EMBED_DIM)))

    transformer_encod = mm.TransformerBlock(
        transformer=BertConfig(hidden_size=EMBED_DIM, num_attention_heads=16),
        pre=mm.ListToDense(max_seq_length=max_seq_length),
    )
    outputs = transformer_encod(inputs)

    if max_seq_length is not None:
        assert list(outputs.shape) == [NUM_ROWS, max_seq_length, EMBED_DIM]
    else:
        assert list(outputs.shape) == [NUM_ROWS, SEQ_LENGTH, EMBED_DIM]


@pytest.mark.parametrize("encoder", [XLNetBlock, BertBlock, AlbertBlock, RobertaBlock, GPT2Block])
def test_hf_tranformers_blocks(encoder):
    NUM_ROWS = 100
    SEQ_LENGTH = 10
    EMBED_DIM = 128
    inputs = tf.random.uniform((NUM_ROWS, SEQ_LENGTH, EMBED_DIM))
    transformer_encod = encoder(
        d_model=EMBED_DIM,
        n_head=8,
        n_layer=2,
    )
    outputs = transformer_encod(inputs)
    assert list(outputs.shape) == [NUM_ROWS, SEQ_LENGTH, EMBED_DIM]


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_transformer_as_classfication_model(sequence_testing_data: Dataset, run_eagerly):
    EMBED_DIM = 48
    loader, schema = classification_loader(sequence_testing_data)

    model = mm.Model(
        mm.InputBlockV2(
            schema,
            embeddings=mm.Embeddings(schema, sequence_combiner=None),
        ),
        BertBlock(
            d_model=EMBED_DIM,
            n_head=8,
            n_layer=2,
            transformer_post="pooler_output",
        ),
        mm.CategoricalOutput(
            to_call=schema["user_country"],
        ),
    )

    batch = next(iter(loader))[0]
    outputs = model(batch)
    assert list(outputs.shape) == [50, 63]
    testing_utils.model_test(model, loader, run_eagerly=run_eagerly)


def test_tranformer_with_prepare_module(sequence_testing_data):
    NUM_ROWS = 100
    SEQ_LENGTH = 10
    EMBED_DIM = 128
    inputs = tf.random.uniform((NUM_ROWS, SEQ_LENGTH, EMBED_DIM))

    class DummyPrepare(tf.keras.layers.Layer):
        def __init__(self, transformer, **kwargs):
            self.transformer = transformer
            super().__init__(**kwargs)

        def call(self, inputs, features=None):
            bs = tf.shape(inputs)[0]
            seq_len = self.transformer.config.max_position_embeddings
            attention_mask = tf.ones((bs, seq_len))
            inputs = {"inputs_embeds": inputs, "attention_mask": attention_mask}
            return inputs

    transformer_encod = BertBlock(
        d_model=EMBED_DIM,
        n_head=8,
        n_layer=2,
        max_position_embeddings=SEQ_LENGTH,
        transformer_pre=DummyPrepare,
    )

    outputs = transformer_encod(inputs)
    assert list(outputs.shape) == [NUM_ROWS, SEQ_LENGTH, EMBED_DIM]


def classification_loader(sequence_testing_data: Dataset):
    def _target_to_onehot(inputs, targets):
        targets = tf.squeeze(tf.one_hot(targets, 63))
        return inputs, targets

    schema = sequence_testing_data.schema.select_by_name(
        ["item_id_seq", "categories", "user_country"]
    )
    schema["user_country"] = schema["user_country"].with_tags(
        schema["user_country"].tags + "target"
    )
    sequence_testing_data.schema = schema
    dataloader = mm.Loader(sequence_testing_data, batch_size=50, transform=_target_to_onehot)
    return dataloader, schema


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_transformer_with_causal_language_modeling(sequence_testing_data: Dataset, run_eagerly):

    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE).select_by_tag(
        Tags.CATEGORICAL
    )
    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]
    predict_next = mm.SequencePredictNext(schema=seq_schema, target=target)

    loader = Loader(sequence_testing_data, batch_size=8, shuffle=False, transform=predict_next)

    model = mm.Model(
        mm.InputBlockV2(
            seq_schema,
            embeddings=mm.Embeddings(
                seq_schema.select_by_tag(Tags.CATEGORICAL), sequence_combiner=None
            ),
        ),
        GPT2Block(d_model=48, n_head=8, n_layer=2),
        mm.CategoricalOutput(
            seq_schema.select_by_name(target), default_loss="categorical_crossentropy"
        ),
    )

    batch = next(iter(loader))[0]
    outputs = model(batch)
    assert list(outputs.shape) == [8, 3, 51997]
    testing_utils.model_test(model, loader, run_eagerly=run_eagerly, reload_model=True)

    metrics = model.evaluate(loader, batch_size=8, steps=1, return_dict=True)
    assert len(metrics) > 0

    predictions = model.predict(loader, batch_size=8, steps=1)
    assert predictions.shape == (8, 3, 51997)


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_transformer_with_masked_language_modeling(sequence_testing_data: Dataset, run_eagerly):

    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE).select_by_tag(
        Tags.CATEGORICAL
    )
    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]

    loader = Loader(sequence_testing_data, batch_size=8, shuffle=False)
    model = mm.Model(
        mm.InputBlockV2(
            seq_schema,
            embeddings=mm.Embeddings(
                seq_schema.select_by_tag(Tags.CATEGORICAL), sequence_combiner=None
            ),
        ),
        # BertBlock(d_model=48, n_head=8, n_layer=2, pre=mm.ReplaceMaskedEmbeddings()),
        GPT2Block(d_model=48, n_head=4, n_layer=2, pre=mm.ReplaceMaskedEmbeddings()),
        mm.CategoricalOutput(
            seq_schema.select_by_name(target),
            default_loss="categorical_crossentropy",
        ),
    )
    seq_mask_random = mm.SequenceMaskRandom(schema=seq_schema, target=target, masking_prob=0.3)

    inputs, targets = next(iter(loader))
    outputs = model(inputs, targets=targets, training=True)
    assert list(outputs.shape) == [8, 4, 51997]
    testing_utils.model_test(
        model,
        loader,
        run_eagerly=run_eagerly,
        reload_model=True,
        fit_kwargs={"pre": seq_mask_random},
    )

    seq_mask_last = mm.SequenceMaskLast(schema=seq_schema, target=target)
    metrics = model.evaluate(loader, batch_size=8, steps=1, return_dict=True, pre=seq_mask_last)
    assert len(metrics) > 0

    predictions = model.predict(loader, batch_size=8, steps=1)
    # TODO: Decide what should be the output of predictions for MLM (currently it predicts for all
    # positions of the sequence, but typically you want a single next-item prediction)
    assert predictions.shape == (8, 4, 51997)


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_transformer_with_masked_language_modeling_check_eval_masked(
    sequence_testing_data: Dataset, run_eagerly
):

    seq_schema = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE).select_by_tag(
        Tags.CATEGORICAL
    )
    target = sequence_testing_data.schema.select_by_tag(Tags.ITEM_ID).column_names[0]

    loader = Loader(sequence_testing_data, batch_size=8, shuffle=False)
    model = mm.Model(
        mm.InputBlockV2(
            seq_schema,
            embeddings=mm.Embeddings(
                seq_schema.select_by_tag(Tags.CATEGORICAL), sequence_combiner=None
            ),
        ),
        # BertBlock(d_model=48, n_head=8, n_layer=2, pre=mm.ReplaceMaskedEmbeddings()),
        GPT2Block(d_model=48, n_head=4, n_layer=2, pre=mm.ReplaceMaskedEmbeddings()),
        mm.CategoricalOutput(
            seq_schema.select_by_name(target),
            default_loss="categorical_crossentropy",
        ),
    )
    seq_mask_random = mm.SequenceMaskRandom(schema=seq_schema, target=target, masking_prob=0.3)

    inputs, targets = next(iter(loader))
    outputs = model(inputs, targets=targets, training=True)
    assert list(outputs.shape) == [8, 4, 51997]
    testing_utils.model_test(
        model,
        loader,
        run_eagerly=run_eagerly,
        reload_model=True,
        fit_kwargs={"pre": seq_mask_random},
        metrics=[mm.RecallAt(5000), mm.NDCGAt(5000)],
    )

    # This transform only extracts targets, but without applying mask
    seq_target_as_input_no_mask = mm.SequenceTargetAsInput(schema=seq_schema, target=target)
    metrics_all_positions1 = model.evaluate(
        loader, batch_size=8, steps=1, return_dict=True, pre=seq_target_as_input_no_mask
    )
    metrics_all_positions2 = model.evaluate(
        loader, batch_size=8, steps=1, return_dict=True, pre=seq_target_as_input_no_mask
    )

    def _metrics_almost_equal(metrics1, metrics2):
        return np.all(
            [
                np.isclose(metrics1[k], metrics2[k])
                for k in metrics1
                if k not in "regularization_loss"
            ]
        )

    # Ensures metrics without masked positions are equal
    assert _metrics_almost_equal(metrics_all_positions1, metrics_all_positions2)

    seq_mask_last = mm.SequenceMaskLast(schema=seq_schema, target=target)
    metrics_last_positions = model.evaluate(
        loader, batch_size=8, steps=1, return_dict=True, pre=seq_mask_last
    )
    # Ensures metrics masking only last positions are different then the ones
    # considering all positions
    assert not _metrics_almost_equal(metrics_all_positions1, metrics_last_positions)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional, Union

import tensorflow as tf
from tensorflow.keras.layers import Layer

from merlin.models.tf.prediction_tasks.base import PredictionTask
from merlin.models.tf.utils.tf_utils import (
    maybe_deserialize_keras_objects,
    maybe_serialize_keras_objects,
)
from merlin.schema import Schema, Tags


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class RegressionTask(PredictionTask):
    """
    Prediction task for regression-task.

    Parameters
    ----------
    target: Union[str, Schema], optional
        The name of the target. If a Schema is provided, the target is inferred from the schema.
    task_name: str, optional
        The name of the task.
    task_block: Block, optional
        The block to use for the task.
    metrics: MetricOrMetrics, optional
        The metrics to use for the task. Defaults to [root-mean-squared-error].
    """

    DEFAULT_LOSS = "mse"
    DEFAULT_METRICS = (tf.keras.metrics.RootMeanSquaredError,)

    def __init__(
        self,
        target: Optional[Union[str, Schema]] = None,
        task_name: Optional[str] = None,
        task_block: Optional[Layer] = None,
        **kwargs,
    ):
        if isinstance(target, Schema):
            target_name = target.select_by_tag(Tags.REGRESSION)
            if not target_name.column_names:
                raise ValueError(
                    "Regression task requires a column with a ", "`Tags.REGRESSION` tag."
                )
            elif len(target_name.column_names) > 1:
                raise ValueError(
                    "Regression task requires a single column with a ",
                    "`Tags.REGRESSION` tag.",
                    "Found {} columns. ".format(len(target_name.column_names)),
                    "Please specify the column name with the `target` argument.",
                )
            target_name = target_name.column_names[0]
        else:
            target_name = target if target else kwargs.pop("target_name", None)

        output_layer = kwargs.pop("output_layer", None)
        super().__init__(
            target_name=target_name,
            task_name=task_name,
            task_block=task_block,
            **kwargs,
        )
        self.output_layer = output_layer or tf.keras.layers.Dense(
            1, name=self.child_name("output_layer")
        )
        # To ensure that the output is always fp32, avoiding numerical
        # instabilities with mixed_float16 policy
        self.output_activation = tf.keras.layers.Activation(
            "linear", dtype="float32", name="prediction"
        )

    def call(self, inputs: tf.Tensor, training=False, **kwargs) -> tf.Tensor:
        """Projects the input with the output layer to a single logit

        Parameters
        ----------
        inputs : tf.Tensor
            Input tensor
        training : bool, optional
            Flag that indicates whether it is training or not, by default False

        Returns
        -------
        tf.Tensor
            Tensor with the regression logit
        """
        return self.output_activation(self.output_layer(inputs))

    def compute_output_shape(self, input_shape):
        return self.output_layer.compute_output_shape(input_shape)

    def get_config(self):
        config = super().get_config()
        config = maybe_serialize_keras_objects(
            self, config, {"output_layer": tf.keras.layers.serialize}
        )

        return config

    @classmethod
    def from_config(cls, config):
        config = maybe_deserialize_keras_objects(
            config, ["output_layer"], tf.keras.layers.deserialize
        )

        return super().from_config(config)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import torch

import merlin.models.torch as ml

if torch.cuda.is_available():
    devices = ["cpu", "cuda"]
else:
    devices = ["cpu"]


def test_simple_model(torch_tabular_features, torch_tabular_data):
    targets = {"target": torch.randint(2, (100,)).float()}

    inputs = torch_tabular_features
    body = ml.SequentialBlock(inputs, ml.MLPBlock([64]))
    model = ml.BinaryClassificationTask("target").to_model(body, inputs)

    dataset = [(torch_tabular_data, targets)]
    losses = model.fit(dataset, num_epochs=5)
    metrics = model.evaluate(dataset, mode="eval")

    # assert list(metrics.keys()) == ["precision", "recall", "accuracy"]
    assert len(metrics) == 3
    assert len(losses) == 5
    assert all(loss.min() >= 0 and loss.max() <= 1 for loss in losses)


# def test_model_with_multiple_heads_and_tasks(
#     yoochoose_schema,
#     torch_yoochoose_tabular_transformer_features,
#     torch_yoochoose_like,
# ):
#     # Tabular classification and regression tasks
#     targets = {
#         "classification": torch.randint(2, (100,)).float(),
#         "regression": torch.randint(2, (100,)).float(),
#     }
#
#     non_sequential_features_schema = yoochoose_schema.select_by_name(["user_age", "user_country"])
#
#     tabular_features = ml.TabularFeatures.from_schema(
#         non_sequential_features_schema,
#         max_sequence_length=20,
#         continuous_projection=64,
#         aggregation="concat",
#     )
#
#     body = ml.SequentialBlock(tabular_features, ml.MLPBlock([64]))
#     tasks = [
#         ml.BinaryClassificationTask("classification"),
#         ml.RegressionTask("regression"),
#     ]
#     head_1 = ml.Head(body, tasks)
#
#     # Session-based classification and regression tasks
#     targets_2 = {
#         "classification_session": torch.randint(2, (100,)).float(),
#         "regression_session": torch.randint(2, (100,)).float(),
#     }
#     transformer_config = tconf.XLNetConfig.build(
#         d_model=64, n_head=4, n_layer=2, total_seq_length=20
#     )
#     body_2 = ml.SequentialBlock(
#         torch_yoochoose_tabular_transformer_features,
#         ml.MLPBlock([64]),
#         ml.TransformerBlock(transformer_config),
#     )
#     tasks_2 = [
#         ml.BinaryClassificationTask("classification_session", summary_type="last"),
#         ml.RegressionTask("regression_session", summary_type="mean"),
#     ]
#     head_2 = ml.Head(body_2, tasks_2)
#
#     # Final model with two heads
#     model = ml.Model(head_1, head_2)
#
#     # launch training
#     targets.update(targets_2)
#     dataset = [(torch_yoochoose_like, targets)]
#     losses = model.fit(dataset, num_epochs=5)
#     metrics = model.evaluate(dataset)
#
#     assert list(metrics.keys()) == [
#         "eval_classification/binary_classification_task",
#         "eval_regression/regression_task",
#         "eval_classification_session/binary_classification_task",
#         "eval_regression_session/regression_task",
#     ]
#     assert len(losses) == 5
#     assert all(loss is not None for loss in losses)


def test_multi_head_model_wrong_weights(torch_tabular_features):
    with pytest.raises(ValueError) as excinfo:
        inputs = torch_tabular_features
        body = ml.SequentialBlock(inputs, ml.MLPBlock([64]))

        head_1 = ml.BinaryClassificationTask("classification").to_head(body, inputs)
        head_2 = ml.RegressionTask("regression").to_head(body, inputs)

        ml.Model(head_1, head_2, head_weights=[0.4])

    assert "`head_weights` needs to have the same length " "as the number of heads" in str(
        excinfo.value
    )



import logging


def require_nvt():
    try:
        import nvtabular as nvt  # noqa

        backend = None
        try:
            import tensorflow as tf

            backend = tf

            if tf.config.list_physical_devices("GPU"):
                _check_nvt_gpu()

        except ImportError:
            pass

        if not backend:
            try:
                import torch

                backend = torch

                if torch.cuda.is_available():
                    _check_nvt_gpu()
            except ImportError:
                pass

    except ImportError:
        raise ImportError(
            "nvtabular is required for this feature.",
            "Please install it with `pip install nvtabular`.",
        )


def _check_nvt_gpu():
    try:
        import cudf  # noqa
    except ImportError:
        logging.warning(
            "A GPU was detected but rapids is not installed.",
            "NVTabular will not be able to use GPU.",
            "Look at the documentation for more information at rapids.ai",
        )



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import pytest

import merlin.models.tf as mm
from merlin.io import Dataset
from merlin.models.tf.utils import testing_utils


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_regression_head(ecommerce_data: Dataset, run_eagerly: bool):
    body = mm.InputBlock(ecommerce_data.schema).connect(mm.MLPBlock([64]))
    model = mm.Model(body, mm.RegressionTask("click"))

    testing_utils.model_test(model, ecommerce_data, run_eagerly=run_eagerly)


@pytest.mark.parametrize("run_eagerly", [True, False])
def test_regression_head_schema(music_streaming_data: Dataset, run_eagerly: bool):
    body = mm.InputBlock(music_streaming_data.schema).connect(mm.MLPBlock([64]))
    model = mm.Model(body, mm.RegressionTask(music_streaming_data.schema))

    testing_utils.model_test(model, music_streaming_data, run_eagerly=run_eagerly)


def test_regression_head_serialization(music_streaming_data: Dataset):
    regression_task = mm.RegressionTask("click")
    assert isinstance(
        regression_task.from_config(regression_task.get_config()), type(regression_task)
    )



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional

import tensorflow as tf

from merlin.models.tf.outputs.sampling.base import Candidate, CandidateSampler


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class PopularityBasedSamplerV2(CandidateSampler):
    """
    Provides a popularity-based negative sampling for the softmax layer
    to ensure training efficiency when the catalog of items is very large.
    The capacity of the queue is fixed and is equal to the catalog size.

    Parameters
    ----------
    max_id: int
        The maximum id value to be sampled. It should be equal to the
        categorical feature cardinality
    min_id: int
        The minimum id value to be sampled. Useful to ignore the first categorical
        encoded ids, which are usually reserved for <nulls>, out-of-vocabulary or padding.
        Defaults to 0.
    max_num_samples: int
        The number of unique negatives to sample at each batch.
    seed: int
        Fix the random values returned by the sampler to ensure reproducibility
        Defaults to None
    """

    def __init__(
        self,
        max_id: int,
        min_id: int = 0,
        max_num_samples: int = 10,
        seed: Optional[int] = None,
        **kwargs,
    ):
        super().__init__(max_num_samples=max_num_samples, **kwargs)
        self.max_id = max_id
        self.min_id = min_id
        self.seed = seed

        assert (
            self.max_num_samples <= self.max_id
        ), f"Number of items to sample `{self.max_num_samples}`"
        f"should be less than total number of ids `{self.max_id}`"

    def add(self, items: Candidate):
        pass

    def call(
        self,
        positive_items: Candidate = None,
        features=None,
        targets=None,
        training=False,
        testing=False,
    ) -> Candidate:
        return self.sample()

    def sample(self) -> Candidate:
        """
        Method to sample `max_num_samples` unique negatives.
        This implementation does not require the actual item frequencies/probabilities
        distribution, but instead tries to approximate the item
        probabilities using the log_uniform (zipfisan) distribution.
        The only requirement is that the item ids are decreasingly sorted by their count frequency.
        We use the default log-uniform (zipfian) sampler given by Tensorflow:
        [log_uniform_candidate_sampler](https://www.tensorflow.org/api_docs/python/tf/random/log_uniform_candidate_sampler)
        We note that the `Categorify` op provided by nvtabular supports
        the frequency-based encoding as default.

        Returns
        -------
        Items
            The negative items ids
        """
        sampled_ids, _, _ = tf.random.log_uniform_candidate_sampler(
            true_classes=tf.ones((1, 1), dtype=tf.int64),
            num_true=1,
            num_sampled=self.max_num_samples,
            unique=True,
            range_max=self.max_id - self.min_id,
            seed=self.seed,
        )

        # Shifting the sampled ids to ignore the first ids (usually reserved for nulls, OOV)
        sampled_ids += self.min_id

        return Candidate(sampled_ids, {})

    def get_config(self):
        config = super().get_config()
        config["max_id"] = self.max_id
        config["min_id"] = self.min_id
        config["max_num_samples"] = self.max_num_samples
        config["seed"] = self.seed
        return config



import os

import pytest

import merlin.io
from merlin.datasets.advertising import get_criteo
from merlin.datasets.synthetic import generate_data

MAYBE_DATA_DIR = os.environ.get("INPUT_DATA_DIR", None)


def test_synthetic_criteo_data():
    dataset = generate_data("criteo", 100)

    assert isinstance(dataset, merlin.io.Dataset)
    assert dataset.num_rows == 100
    assert len(dataset.schema) == 40


@pytest.mark.skipif(
    MAYBE_DATA_DIR is None,
    reason="No data-dir available, pass it through env variable $INPUT_DATA_DIR",
)
def test_get_criteo(tmp_path):
    data_path = os.path.join(MAYBE_DATA_DIR, "criteo")

    train, valid = get_criteo(data_path, num_days=2)

    assert isinstance(train, merlin.io.Dataset)
    assert isinstance(valid, merlin.io.Dataset)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pandas as pd
import pytest
import tensorflow as tf

import merlin.models.tf as mm
from merlin.io import Dataset
from merlin.models.tf.core.prediction import Prediction
from merlin.models.tf.transforms.negative_sampling import InBatchNegatives
from merlin.models.tf.utils import testing_utils
from merlin.schema import ColumnSchema, Schema, Tags


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class ExampleIsTraining(tf.keras.layers.Layer):
    def call(self, inputs, training=False):
        return training


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class ExamplePredictionIdentity(tf.keras.layers.Layer):
    def call(self, inputs, targets=None):
        return Prediction(inputs, targets)

    def compute_output_shape(self, input_shape):
        return input_shape


class TestAddRandomNegativesToBatch:
    def test_dataloader(self):
        schema = Schema(
            [
                ColumnSchema(
                    "item_id", tags=[Tags.ITEM, Tags.ITEM_ID, Tags.CATEGORICAL], dtype="int64"
                ),
                ColumnSchema("item_feature", tags=[Tags.ITEM, Tags.CATEGORICAL]),
                ColumnSchema(
                    "user_id", tags=[Tags.USER, Tags.USER_ID, Tags.CATEGORICAL], dtype="int64"
                ),
                ColumnSchema("user_feature", tags=[Tags.USER, Tags.CATEGORICAL]),
                ColumnSchema("label", tags=[Tags.TARGET]),
            ]
        )
        n_per_positive = 5
        sampler = InBatchNegatives(schema, n_per_positive)

        input_df = pd.DataFrame(
            [
                {"user_id": 1, "item_id": 1, "item_feature": 2, "user_feature": 10, "label": 1},
                {"user_id": 2, "item_id": 3, "item_feature": 6, "user_feature": 5, "label": 1},
            ]
        )
        input_df = input_df[sorted(input_df.columns)]
        dataset = Dataset(input_df, schema=schema)
        loader = mm.Loader(dataset, batch_size=10, transform=sampler)
        outputs, targets = next(iter(loader))

        output_dict = {
            key: output_tensor.numpy().reshape(-1) for key, output_tensor in outputs.items()
        }
        output_df = pd.DataFrame({**output_dict, "label": targets.numpy().reshape(-1)})
        output_df = output_df[sorted(output_df.columns)]

        # first part of outputs frame should match inputs
        pd.testing.assert_frame_equal(input_df, output_df[: len(input_df)])

        # negatives added to batch should not overlap with positive user-items

        def user_item_pairs(df):
            return {tuple(x) for x in df[["user_id", "item_id"]].values.tolist()}

        input_user_item_pairs = user_item_pairs(input_df)
        negatives_df = output_df[len(input_df) :]
        negatives_user_item_pairs = user_item_pairs(negatives_df)

        assert input_user_item_pairs.intersection(negatives_user_item_pairs) == set()

    def assert_outputs_batch_size(self, assert_fn, *outputs):
        for values in zip(*outputs):
            for value in values:
                if isinstance(value, tuple):
                    assert_fn(value[1].shape[0])
                else:
                    assert_fn(value.shape[0])

    @pytest.mark.parametrize("to_dense", [True, False])
    def test_calling_without_targets(
        self, music_streaming_data: Dataset, to_dense: bool, tf_random_seed: int
    ):
        schema = music_streaming_data.schema
        batch_size, n_per_positive = 10, 5
        features = mm.sample_batch(
            music_streaming_data, batch_size=batch_size, include_targets=False, to_dense=to_dense
        )

        sampler = InBatchNegatives(schema, n_per_positive, seed=tf_random_seed)

        outputs = sampler(features).outputs

        def assert_fn(output_batch_size):
            assert output_batch_size == batch_size

        self.assert_outputs_batch_size(assert_fn, outputs.values())

    @pytest.mark.parametrize("to_dense", [True, False])
    def test_calling(self, music_streaming_data: Dataset, to_dense: bool, tf_random_seed: int):
        schema = music_streaming_data.schema
        batch_size, n_per_positive = 10, 5
        inputs, targets = mm.sample_batch(
            music_streaming_data, batch_size=batch_size, include_targets=True, to_dense=to_dense
        )

        sampler = InBatchNegatives(schema, 5, seed=tf_random_seed)

        outputs, targets = sampler(inputs, targets=targets)

        max_batch_size = batch_size + batch_size * n_per_positive

        def assert_fn(output_batch_size):
            assert batch_size < output_batch_size <= max_batch_size

        self.assert_outputs_batch_size(
            assert_fn,
            outputs.values(),
            targets.values(),
        )

    @pytest.mark.parametrize("to_dense", [True, False])
    def test_run_when_testing(
        self, music_streaming_data: Dataset, to_dense: bool, tf_random_seed: int
    ):
        schema = music_streaming_data.schema
        batch_size, n_per_positive = 10, 5
        inputs, targets = mm.sample_batch(
            music_streaming_data, batch_size=batch_size, include_targets=True, to_dense=to_dense
        )

        sampler = InBatchNegatives(
            schema, n_per_positive, seed=tf_random_seed, run_when_testing=False
        )

        with_negatives = sampler(inputs, targets=targets, testing=True)
        outputs = with_negatives.outputs
        targets = with_negatives.targets

        def assert_fn(output_batch_size):
            assert output_batch_size == batch_size

        self.assert_outputs_batch_size(
            assert_fn,
            outputs.values(),
            targets.values(),
        )

    # The sampling layer currnetly only works correctly as part of the model when run in eager mode
    @pytest.mark.parametrize("run_eagerly", [True])
    def test_in_model(self, run_eagerly, music_streaming_data: Dataset, tf_random_seed: int):
        dataset = music_streaming_data
        schema = dataset.schema

        sampling = InBatchNegatives(schema, 5, seed=tf_random_seed)

        model = mm.Model(
            mm.InputBlock(schema),
            sampling,
            mm.MLPBlock([64]),
            mm.BinaryClassificationTask("click"),
        )

        batch_size = 10
        features, targets = mm.sample_batch(dataset, batch_size=batch_size, to_ragged=True)

        with_negatives = model(features, targets=targets, training=True)
        assert with_negatives.predictions.shape[0] >= 50

        without_negatives = model(features)
        assert without_negatives.shape[0] == batch_size

        preds = model.predict(features)
        assert preds.shape[0] == batch_size

        testing_utils.model_test(model, dataset, run_eagerly=run_eagerly, reload_model=True)

    def test_model_with_dataloader(self, music_streaming_data: Dataset, tf_random_seed: int):
        dataset = music_streaming_data
        schema = dataset.schema

        add_negatives = InBatchNegatives(schema, 5, seed=tf_random_seed)

        batch_size, n_per_positive = 10, 5
        loader = mm.Loader(dataset, batch_size=batch_size, transform=add_negatives)

        features, targets = next(iter(loader))

        expected_batch_size = batch_size + batch_size * n_per_positive

        assert features["item_category"].shape[0] > batch_size
        assert features["item_category"].shape[0] <= expected_batch_size
        assert all(
            f.shape[0] > batch_size and f.shape[0] <= expected_batch_size for f in features.values()
        )
        assert all(
            f.shape[0] > batch_size and f.shape[0] <= expected_batch_size for f in targets.values()
        )

        model = mm.Model(
            mm.InputBlock(schema),
            mm.MLPBlock([64]),
            mm.BinaryClassificationTask("click"),
        )
        assert model(features).shape[0] > batch_size
        assert model(features).shape[0] <= expected_batch_size

        testing_utils.model_test(model, loader)



# Test is currently breaks in TF 2.10

from testbook import testbook

from tests.conftest import REPO_ROOT

p = "examples/usecases/accelerate-training-of-large-embedding-tables-by-LazyAdam.ipynb"


@testbook(
    REPO_ROOT / p,
    timeout=180,
    execute=False,
)
def test_usecase_accelerate_training_by_lazyadam(tb):
    tb.inject(
        """
        import os
        os.environ["NUM_ROWS"] = "1000"
        """
    )
    tb.execute()
    model1_lazyadam = tb.ref("model1")
    model2_adam = tb.ref("model2")
    assert set(model1_lazyadam.history.history.keys()) == set(
        [
            "loss",
            "recall_at_10",
            "mrr_at_10",
            "ndcg_at_10",
            "map_at_10",
            "precision_at_10",
            "regularization_loss",
        ]
    )
    assert set(model2_adam.history.history.keys()) == set(
        [
            "loss",
            "recall_at_10",
            "mrr_at_10",
            "ndcg_at_10",
            "map_at_10",
            "precision_at_10",
            "regularization_loss",
        ]
    )



import os

import pytest

import merlin.io
from merlin.datasets import ecommerce
from merlin.datasets.synthetic import generate_data

# This is the path to the raw ali-ccp dataset
MAYBE_ALICCP_DATA = os.environ.get("DATA_PATH_ALICCP", None)


def test_synthetic_aliccp_data():
    dataset = generate_data("aliccp", 100)

    assert isinstance(dataset, merlin.io.Dataset)
    assert dataset.num_rows == 100
    assert len(dataset.schema) == 27
    assert dataset.compute()["click"].sum() > 0


def test_synthetic_aliccp_raw_data(tmp_path):
    dataset = generate_data("aliccp-raw", 100)

    assert isinstance(dataset, merlin.io.Dataset)
    assert dataset.num_rows == 100
    assert len(dataset.schema) == 25
    assert sorted(dataset.to_ddf().compute().columns) == [
        "click",
        "conversion",
        "item_brand",
        "item_category",
        "item_id",
        "item_intention",
        "item_shop",
        "position",
        "user_age",
        "user_brands",
        "user_categories",
        "user_consumption_1",
        "user_consumption_2",
        "user_gender",
        "user_geography",
        "user_group",
        "user_id",
        "user_intentions",
        "user_is_occupied",
        "user_item_brands",
        "user_item_categories",
        "user_item_intentions",
        "user_item_shops",
        "user_profile",
        "user_shops",
    ]

    ecommerce.transform_aliccp((dataset, dataset), tmp_path)
    output_files = list(tmp_path.glob("*/*"))

    assert len(output_files) == 10


@pytest.mark.skipif(
    MAYBE_ALICCP_DATA is None,
    reason="ALI-CCP data is not available, pass it through env variable $DATA_PATH_ALICCP",
)
def test_get_alliccp():
    data_path = MAYBE_ALICCP_DATA

    nvt_workflow = ecommerce.default_aliccp_transformation(add_target_encoding=False)
    train, valid = ecommerce.get_aliccp(
        data_path, nvt_workflow=nvt_workflow, transformed_name="raw_transform", overwrite=True
    )

    assert isinstance(train, merlin.io.Dataset)
    assert isinstance(valid, merlin.io.Dataset)


@pytest.mark.skipif(
    MAYBE_ALICCP_DATA is None,
    reason="ALI-CCP data is not available, pass it through env variable $DATA_PATH_ALICCP",
)
def test_prepare_alliccp(tmp_path):
    data_path = MAYBE_ALICCP_DATA

    ecommerce.prepare_alliccp(data_path, file_size=50, max_num_rows=100, output_dir=tmp_path)
    output_files = list(tmp_path.glob("*/*"))

    assert len(output_files) == 2
    assert all(f.name.endswith(".parquet") for f in output_files)


@pytest.mark.skipif(
    MAYBE_ALICCP_DATA is None,
    reason="ALI-CCP data is not available, pass it through env variable $DATA_PATH_ALICCP",
)
def test_transform_alliccp(tmp_path):
    data_path = MAYBE_ALICCP_DATA

    ecommerce.transform_aliccp(data_path, tmp_path)
    output_files = list(tmp_path.glob("*/*"))

    assert len(output_files) == 10



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


import pytest
import tensorflow as tf

import merlin.models.tf as ml
from merlin.io import Dataset
from merlin.models.tf.utils import testing_utils
from merlin.schema import ColumnSchema, Schema, Tags


def test_tabular_features(testing_data: Dataset):
    tab_module = ml.InputBlock(testing_data.schema)

    outputs = tab_module(ml.sample_batch(testing_data, batch_size=100, include_targets=False))

    con = testing_data.schema.select_by_tag(Tags.CONTINUOUS).column_names
    cat = testing_data.schema.select_by_tag(Tags.CATEGORICAL).column_names

    assert set(outputs.keys()) == set(con + cat)


def test_serialization_tabular_features(testing_data: Dataset):
    inputs = ml.InputBlock(testing_data.schema)

    copy_layer = testing_utils.assert_serialization(inputs)

    assert list(inputs.parallel_layers.keys()) == list(copy_layer.parallel_layers.keys())


def test_tabular_features_with_projection(testing_data: Dataset):
    tab_module = ml.InputBlock(testing_data.schema, continuous_projection=ml.MLPBlock([64]))

    outputs = tab_module(ml.sample_batch(testing_data, batch_size=100, include_targets=False))
    continuous_feature_names = testing_data.schema.select_by_tag(Tags.CONTINUOUS).column_names

    assert len(set(continuous_feature_names).intersection(set(outputs.keys()))) == 0
    assert "continuous_projection" in outputs
    assert list(outputs["continuous_projection"].shape)[1] == 64


@testing_utils.mark_run_eagerly_modes
@pytest.mark.parametrize("continuous_projection", [None, 128])
def test_tabular_features_yoochoose_model(
    music_streaming_data: Dataset, run_eagerly, continuous_projection
):
    if continuous_projection:
        continuous_projection = ml.MLPBlock([continuous_projection])
    inputs = ml.InputBlock(
        music_streaming_data.schema,
        continuous_projection=continuous_projection,
        aggregation="concat",
    )

    body = ml.SequentialBlock([inputs, ml.MLPBlock([64])])
    model = ml.Model(body, ml.BinaryClassificationTask("click"))

    testing_utils.model_test(model, music_streaming_data, run_eagerly=run_eagerly)


@testing_utils.mark_run_eagerly_modes
@pytest.mark.parametrize("continuous_projection", [None, 128])
def test_tabular_features_yoochoose_model_inputblockv2(
    music_streaming_data: Dataset, run_eagerly, continuous_projection
):
    kwargs = {}
    if continuous_projection:
        kwargs["continuous"] = ml.ContinuousProjection(
            music_streaming_data.schema.select_by_tag(Tags.CONTINUOUS),
            ml.MLPBlock([continuous_projection]),
        )

    inputs = ml.InputBlockV2(music_streaming_data.schema, aggregation="concat", **kwargs)

    body = ml.SequentialBlock([inputs, ml.MLPBlock([64])])
    model = ml.Model(body, ml.BinaryClassificationTask("click"))

    testing_utils.model_test(model, music_streaming_data, run_eagerly=run_eagerly)


def test_tabular_seq_features_ragged_embeddings(sequence_testing_data: Dataset):
    tab_module = ml.InputBlockV2(
        sequence_testing_data.schema,
        categorical=ml.Embeddings(
            sequence_testing_data.schema.select_by_tag(Tags.CATEGORICAL), sequence_combiner=None
        ),
        aggregation=None,
    )

    batch = ml.sample_batch(
        sequence_testing_data, batch_size=100, include_targets=False, to_ragged=True
    )

    outputs = tab_module(batch)

    con = sequence_testing_data.schema.select_by_tag(Tags.CONTINUOUS).column_names
    cat = sequence_testing_data.schema.select_by_tag(Tags.CATEGORICAL).column_names
    seq = sequence_testing_data.schema.select_by_tag(Tags.SEQUENCE).column_names

    assert set(outputs.keys()) == set(con + cat)
    assert all(isinstance(val, tf.RaggedTensor) for name, val in outputs.items() if name in seq)


@pytest.mark.parametrize(
    "seq_combiner",
    [tf.keras.layers.Lambda(lambda x: tf.reduce_mean(x, axis=1)), "mean"],
)
def test_tabular_seq_features_ragged_emb_combiner(sequence_testing_data: Dataset, seq_combiner):
    con2d = sequence_testing_data.schema.select_by_tag(Tags.CONTINUOUS).remove_by_tag(Tags.SEQUENCE)
    input_block = ml.InputBlockV2(
        sequence_testing_data.schema,
        categorical=ml.Embeddings(
            sequence_testing_data.schema.select_by_tag(Tags.CATEGORICAL),
            sequence_combiner=seq_combiner,
        ),
        continuous=ml.Continuous(con2d),
        aggregation=None,
    )

    batch = ml.sample_batch(
        sequence_testing_data, batch_size=100, include_targets=False, to_ragged=True
    )

    outputs = input_block(batch)

    cat = sequence_testing_data.schema.select_by_tag(Tags.CATEGORICAL).column_names

    assert all(isinstance(val, tf.Tensor) for name, val in outputs.items())
    assert all(tf.rank(val) == 2 for name, val in outputs.items() if name in cat)
    assert set(cat + con2d.column_names) == set(outputs.keys())


def test_tabular_seq_features_ragged_custom_emb_combiner(sequence_testing_data: Dataset):
    schema = sequence_testing_data.schema
    schema = schema + Schema([ColumnSchema("item_id_seq_weights")])
    assert "item_id_seq_weights" in schema.column_names

    batch = ml.sample_batch(
        sequence_testing_data, batch_size=100, include_targets=False, to_ragged=True
    )
    batch["item_id_seq_weights"] = tf.ragged.constant(
        [[1.0, 2.0, 3.0, 4.0] for _ in range(batch["item_id_seq"].shape[0])],
        row_splits_dtype=batch["item_id_seq"].row_splits.dtype,
    )

    input_block_weighed_avg = ml.InputBlockV2(
        schema,
        categorical=ml.Embeddings(
            schema.select_by_tag(Tags.CATEGORICAL),
            sequence_combiner=ml.AverageEmbeddingsByWeightFeature.from_schema_convention(
                schema, "_weights"
            ),
        ),
        aggregation=None,
    )

    outputs_weighted_avg = input_block_weighed_avg(batch, features=batch)

    input_block_simple_avg = ml.InputBlockV2(
        schema,
        categorical=ml.Embeddings(
            schema.select_by_tag(Tags.CATEGORICAL),
            sequence_combiner=tf.keras.layers.Lambda(lambda x: tf.reduce_mean(x, axis=1)),
        ),
        aggregation=None,
    )

    outputs_simple_avg = input_block_simple_avg(batch, features=batch)

    assert not tf.reduce_all(
        outputs_weighted_avg["item_id_seq"] == outputs_simple_avg["item_id_seq"]
    )

    cat = schema.select_by_tag(Tags.CATEGORICAL).column_names

    assert all(
        isinstance(val, tf.Tensor) for name, val in outputs_weighted_avg.items() if name in cat
    )
    assert all(tf.rank(val) == 2 for name, val in outputs_weighted_avg.items() if name in cat)


def test_tabular_seq_features_avg_embeddings_with_mapvalues(sequence_testing_data: Dataset):
    cat_schema = sequence_testing_data.schema.select_by_tag(Tags.CATEGORICAL)

    batch = ml.sample_batch(
        sequence_testing_data, batch_size=100, include_targets=False, to_ragged=True
    )

    input_block = ml.InputBlockV2(
        cat_schema,
        categorical=ml.Embeddings(
            cat_schema,
        ),
        post=ml.MapValues(
            tf.keras.layers.Lambda(
                lambda x: tf.math.reduce_mean(x, axis=1) if isinstance(x, tf.RaggedTensor) else x
            )
        ),
        aggregation=None,
    )

    output = input_block(batch)
    assert all(
        isinstance(val, tf.Tensor)
        for name, val in output.items()
        if name in cat_schema.column_names
    )
    assert all(tf.rank(val) == 2 for name, val in output.items() if name in cat_schema.column_names)


@pytest.mark.parametrize("aggregation", [None, "concat"])
def test_embedding_tables_from_schema_infer_dims(sequence_testing_data: Dataset, aggregation: str):
    cat_schema = sequence_testing_data.schema.select_by_tag(Tags.CATEGORICAL)
    embeddings_block = ml.Embeddings(
        cat_schema.select_by_tag(Tags.CATEGORICAL),
        dim={"item_id_seq": 15, "test_user_id": 21},
        embeddings_initializer="truncated_normal",
    )
    input_block = ml.InputBlockV2(cat_schema, categorical=embeddings_block, aggregation=aggregation)

    batch = ml.sample_batch(
        sequence_testing_data, batch_size=100, include_targets=False, to_ragged=True
    )

    outputs = input_block(batch)

    if aggregation == "concat":
        assert outputs.shape[-1] == 60
    elif aggregation is None:
        assert {k: v.shape[-1] for k, v in outputs.items()} == {
            "test_user_id": 21,
            "item_id_seq": 15,
            # Inferred dims from cardinality
            "categories": 16,
            "user_country": 8,
        }



#
# Copyright (c) 2022, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pandas as pd
import pytest
import sklearn.datasets
import xgboost

from merlin.core.dispatch import HAS_GPU
from merlin.datasets.synthetic import generate_data
from merlin.io import Dataset
from merlin.models.xgb import XGBoost, dataset_to_xy


def test_without_dask_client(music_streaming_data: Dataset):
    with pytest.raises(ValueError) as exc_info:
        model = XGBoost(music_streaming_data.schema, objective="reg:logistic")
        model.fit(music_streaming_data)
    assert "No global client found" in str(exc_info.value)


@pytest.mark.usefixtures("dask_client")
class TestXGBoost:
    def test_unsupported_objective(self, music_streaming_data: Dataset):
        with pytest.raises(ValueError) as excinfo:
            model = XGBoost(music_streaming_data.schema, objective="reg:unknown")
            model.fit(music_streaming_data)
        assert "Objective not supported" in str(excinfo.value)

    def test_music_regression(self, music_streaming_data: Dataset):
        schema = music_streaming_data.schema
        model = XGBoost(schema, objective="reg:logistic")
        model.fit(music_streaming_data)
        model.predict(music_streaming_data)
        metrics = model.evaluate(music_streaming_data)

        assert "rmse" in metrics

    def test_ecommerce_click(self, ecommerce_data: Dataset):
        schema = ecommerce_data.schema
        model = XGBoost(
            schema, target_columns=["click"], objective="binary:logistic", eval_metric="auc"
        )
        model.fit(ecommerce_data)
        model.predict(ecommerce_data)
        metrics = model.evaluate(ecommerce_data)

        assert "auc" in metrics

    def test_social_click(self, social_data: Dataset):
        schema = social_data.schema
        model = XGBoost(
            schema, target_columns=["click"], objective="binary:logistic", eval_metric=["auc"]
        )
        model.fit(social_data)
        model.predict(social_data)
        metrics = model.evaluate(social_data)

        assert "auc" in metrics

    def test_logistic(self, criteo_data: Dataset):
        schema = criteo_data.schema
        model = XGBoost(schema, objective="binary:logistic", eval_metric=["auc"])
        model.fit(criteo_data)
        model.predict(criteo_data)
        metrics = model.evaluate(criteo_data)

        assert "auc" in metrics

    def test_ndcg(self, social_data: Dataset):
        schema = social_data.schema
        model = XGBoost(
            schema,
            target_columns="click",
            qid_column="user_id",
            objective="rank:ndcg",
            eval_metric=["auc", "ndcg", "map"],
        )
        model.fit(social_data)
        model.predict(social_data)
        metrics = model.evaluate(social_data)

        assert "map" in metrics

    def test_pairwise(self, social_data: Dataset):
        schema = social_data.schema
        model = XGBoost(
            schema,
            target_columns=["click"],
            qid_column="user_id",
            objective="rank:pairwise",
            eval_metric=["ndcg", "auc", "map"],
        )
        model.fit(social_data)
        model.predict(social_data)
        model.evaluate(social_data)


@pytest.mark.skipif(not HAS_GPU, reason="No GPU available")
@pytest.mark.parametrize(
    ["fit_kwargs", "expected_dtrain_cls"],
    [
        ({}, xgboost.dask.DaskDeviceQuantileDMatrix),
        ({"use_quantile": False}, xgboost.dask.DaskDMatrix),
    ],
)
@patch("xgboost.dask.train", side_effect=xgboost.dask.train)
def test_gpu_hist_dmatrix(
    mock_train, fit_kwargs, expected_dtrain_cls, dask_client, music_streaming_data: Dataset
):
    schema = music_streaming_data.schema
    model = XGBoost(schema, objective="reg:logistic", tree_method="gpu_hist")
    model.fit(music_streaming_data, **fit_kwargs)
    model.predict(music_streaming_data)
    metrics = model.evaluate(music_streaming_data)
    assert "rmse" in metrics

    assert mock_train.called
    assert mock_train.call_count == 1

    train_call = mock_train.call_args_list[0]
    client, params, dtrain = train_call.args
    assert dask_client == client
    assert params["tree_method"] == "gpu_hist"
    assert params["objective"] == "reg:logistic"
    assert isinstance(dtrain, expected_dtrain_cls)


@pytest.mark.usefixtures("dask_client")
class TestSchema:
    def test_fit_with_sub_schema(self, music_streaming_data: Dataset):
        schema = music_streaming_data.schema
        sub_schema = schema.select_by_name(["session_id", "country", "play_percentage"])
        model = XGBoost(sub_schema, objective="reg:logistic")
        model.fit(music_streaming_data)
        assert model.booster.num_features() == 2

    def test_no_features(self, music_streaming_data: Dataset):
        schema = music_streaming_data.schema
        sub_schema = schema.select_by_name(["unknown_feature", "play_percentage"])
        with pytest.raises(ValueError) as excinfo:
            XGBoost(sub_schema, objective="reg:logistic")
        assert "No feature columns found" in str(excinfo.value)

    def test_fit_with_missing_features(self, music_streaming_data: Dataset):
        schema = music_streaming_data.schema
        sub_schema = schema.select_by_name(["session_id", "play_percentage"])
        model = XGBoost(sub_schema, objective="reg:logistic")
        df = music_streaming_data.to_ddf().compute()
        new_dataset = Dataset(df[["click", "play_percentage"]])
        with pytest.raises(KeyError) as excinfo:
            model.fit(new_dataset)
        assert "session_id" in str(excinfo)


class TestEvals:
    def test_multiple(self, dask_client):
        train, valid_a, valid_b = generate_data(
            "music-streaming", num_rows=100, set_sizes=(0.6, 0.2, 0.2)
        )
        model = XGBoost(train.schema, objective="reg:logistic")
        model.fit(train, evals=[(valid_a, "a"), (valid_b, "b")])
        assert set(model.evals_result.keys()) == {"a", "b"}

    def test_default(self, dask_client):
        train = generate_data("music-streaming", num_rows=100)
        model = XGBoost(train.schema, objective="reg:logistic")
        model.fit(train)
        assert set(model.evals_result.keys()) == {"train"}

    def test_train_and_valid(self, dask_client):
        train, valid = generate_data("music-streaming", num_rows=100, set_sizes=(0.5, 0.5))
        model = XGBoost(train.schema, objective="reg:logistic")
        model.fit(train, evals=[(valid, "valid"), (train, "train")])
        assert set(model.evals_result.keys()) == {"valid", "train"}

    def test_invalid_data(self, dask_client):
        train, _ = generate_data("music-streaming", num_rows=100, set_sizes=(0.5, 0.5))
        model = XGBoost(train.schema, objective="reg:logistic")
        with pytest.raises(AssertionError):
            model.fit(train, evals=[([], "valid")])


def test_dataset_to_xy_does_not_modify_column_order():
    df = pd.DataFrame(data={"z": [0], "target": [-1], "a": [1], "Z": [2]})
    feature_columns = ["z", "Z", "a"]
    X, y, _ = dataset_to_xy(
        dataset=Dataset(df),
        feature_columns=feature_columns,
        target_columns="target",
        qid_column=None,
    )
    assert X.columns.tolist() == feature_columns


def test_predict_without_target(dask_client):
    rows = 200
    num_features = 16
    X, y = sklearn.datasets.make_regression(
        n_samples=rows,
        n_features=num_features,
        n_informative=num_features // 3,
        random_state=0,
    )

    feature_names = [str(i) for i in range(num_features)]
    df = pd.DataFrame(
        np.hstack((X, y.reshape(-1, 1))), columns=feature_names + ["target"], dtype=np.float32
    )
    train_ds = Dataset(df.iloc[:160])
    valid_ds = Dataset(df.iloc[40:])
    test_ds = Dataset(df.iloc[40:].drop(columns="target"))

    model = XGBoost(schema=train_ds.schema, target_columns="target")
    model.fit(train_ds, evals=[(valid_ds, "validation_set")])
    model.predict(test_ds)


def test_reload(dask_client, tmpdir):
    train, valid = generate_data("social", num_rows=40, set_sizes=(0.5, 0.5))

    schema = train.schema
    xgb_booster_params = {
        "objective": "rank:pairwise",
    }

    xgb_train_params = {
        "num_boost_round": 1,
        "verbose_eval": 1,
        "early_stopping_rounds": 1,
    }

    model = XGBoost(schema, target_columns=["click"], qid_column="user_id", **xgb_booster_params)
    model.fit(
        train,
        evals=[
            (valid, "validation_set"),
        ],
        **xgb_train_params
    )
    _ = model.evaluate(valid)

    model_dir = Path(tmpdir) / "xgb_model"

    model.save(model_dir)
    reloaded = XGBoost.load(model_dir)

    np.testing.assert_array_almost_equal(model.predict(valid), reloaded.predict(valid))

    assert reloaded.schema == model.schema
    assert reloaded.target_columns == model.target_columns
    assert reloaded.feature_columns == model.feature_columns
    assert reloaded.qid_column == model.qid_column



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional, Sequence, Type, Union

import tensorflow as tf
from tensorflow.keras.layers import Layer

from merlin.models.config.schema import SchemaMixin
from merlin.models.tf.typing import TabularData
from merlin.models.utils.registry import Registry
from merlin.schema import Schema, Tags

block_registry: Registry = Registry.class_registry("tf.blocks")
BlockType = Union["Block", str, Sequence[str]]


if TYPE_CHECKING:
    from merlin.models.tf.core.combinators import (
        Filter,
        ParallelBlock,
        SequentialBlock,
        TabularAggregationType,
    )
    from merlin.models.tf.prediction_tasks.base import PredictionTask


class PredictionOutput(NamedTuple):
    "class to hold PredictionOutput"

    predictions: Union[TabularData, tf.Tensor]
    targets: Union[TabularData, tf.Tensor]
    positive_item_ids: Optional[tf.Tensor] = None
    label_relevant_counts: Optional[tf.Tensor] = None
    valid_negatives_mask: Optional[tf.Tensor] = None
    negative_item_ids: Optional[tf.Tensor] = None
    sample_weight: Optional[tf.Tensor] = None

    def copy_with_updates(
        self,
        predictions: Optional[Union[TabularData, tf.Tensor]] = None,
        targets: Optional[Union[TabularData, tf.Tensor]] = None,
        positive_item_ids: Optional[tf.Tensor] = None,
        label_relevant_counts: Optional[tf.Tensor] = None,
        valid_negatives_mask: Optional[tf.Tensor] = None,
        negative_item_ids: Optional[tf.Tensor] = None,
        sample_weight: Optional[tf.Tensor] = None,
    ):
        """Creates a new instance of PredictionOutput
        allowing to override the attributes for the copy
        """
        output = PredictionOutput(
            predictions=(self.predictions if predictions is None else predictions),
            targets=(self.targets if targets is None else targets),
            positive_item_ids=(
                self.positive_item_ids if positive_item_ids is None else positive_item_ids
            ),
            label_relevant_counts=(
                self.label_relevant_counts
                if label_relevant_counts is None
                else label_relevant_counts
            ),
            valid_negatives_mask=(
                self.valid_negatives_mask if valid_negatives_mask is None else valid_negatives_mask
            ),
            negative_item_ids=(
                self.negative_item_ids if negative_item_ids is None else negative_item_ids
            ),
            sample_weight=(self.sample_weight if sample_weight is None else sample_weight),
        )
        return output


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class ModelContext(Layer):
    """ModelContext is used to store/retrieve public variables across blocks.

    (This is created automatically in the model and doesn't need to be created manually.)
    """

    def add_variable(self, variable):
        setattr(self, variable.name, variable)

    def add_embedding_table(self, name, embedding_table):
        embedding_tables = getattr(self, "embedding_tables", {})
        embedding_tables[name] = embedding_table
        setattr(self, "embedding_tables", embedding_tables)

    def set_dtypes(self, features):
        for feature_name in features:
            feature = features[feature_name]

            if isinstance(feature, tuple):
                dtype = feature[0].dtype
            else:
                dtype = feature.dtype

            self._feature_dtypes[feature_name] = dtype

    def __getitem__(self, item):
        if isinstance(item, Schema):
            if len(item.column_names) > 1:
                raise ValueError("Schema contains more than one column.")
            item = item.column_names[0]
        elif isinstance(item, Tags):
            item = item.value
        else:
            item = str(item)
        return self.named_variables[item]

    def get_embedding(self, item):
        if isinstance(item, Tags):
            item = item.value
        else:
            item = str(item)

        embedding_tables = getattr(self, "embedding_tables", {})
        return embedding_tables[item].embeddings

    @property
    def named_variables(self) -> Dict[str, tf.Variable]:
        outputs = {}
        for var in self.variables:
            if var.name.endswith("/embedding:0"):
                name = "/".join(var.name.split("/")[-2:])
            else:
                name = var.name.split("/")[-1]
            outputs[name.replace(":0", "")] = var

        return outputs

    def get_config(self):
        config = super(ModelContext, self).get_config()

        return config


class ContextMixin:
    @property
    def context(self) -> ModelContext:
        return self._context

    def _set_context(self, context: ModelContext):
        self._context: ModelContext = context


class Block(SchemaMixin, ContextMixin, Layer):
    """Core abstraction in Merlin models."""

    registry = block_registry

    def __init__(self, context: Optional[ModelContext] = None, **kwargs):
        super(Block, self).__init__(**kwargs)
        if context:
            self._set_context(context)

    @classmethod
    @tf.autograph.experimental.do_not_convert
    def parse(cls, *block: BlockType) -> "Block":
        if len(block) == 1 and isinstance(block[0], (list, tuple)):
            block = block[0]  # type: ignore

        output: "Block"
        if len(block) == 1:
            output = cls.registry.parse(block[0])
        else:
            blocks = [cls.registry.parse(b) for b in block]
            output = blocks[0].connect(*blocks[1:])

        return output

    @classmethod
    def from_layer(cls, layer: tf.keras.layers.Layer) -> "Block":
        layer.__class__ = cls

        return layer  # type: ignore

    @classmethod
    def parse_block(cls, input: Union["Block", tf.keras.layers.Layer]) -> "Block":
        if isinstance(input, Block):
            return input

        return cls.from_layer(input)

    def build(self, input_shapes):
        self._maybe_propagate_context(input_shapes)

        return super().build(input_shapes)

    def call_outputs(
        self, outputs: PredictionOutput, training=False, **kwargs
    ) -> "PredictionOutput":
        return outputs

    def register_features(self, feature_shapes) -> List[str]:
        return []

    def as_tabular(self, name=None) -> "Block":
        from merlin.models.tf.core.combinators import SequentialBlock
        from merlin.models.tf.core.tabular import AsTabular

        if not name:
            name = self.name

        return SequentialBlock([self, AsTabular(name)], copy_layers=False)

    def repeat(self, num: int = 1) -> "SequentialBlock":
        """Repeat the block num times.

        Parameters
        ----------
        num : int
            Number of times to repeat the block.
        """
        from merlin.models.tf.core.combinators import SequentialBlock

        repeated = []
        for _ in range(num):
            repeated.append(self.copy())

        return SequentialBlock(repeated)

    def prepare(
        self,
        block: Optional[BlockType] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional["TabularAggregationType"] = None,
    ) -> "SequentialBlock":
        """Transform the inputs of this block.

        Parameters
        ----------
        block: Optional[Block]
            If set, this block will be used to transform the inputs of this block.
        post: Block
            Block to use as post-transformation.
        aggregation: TabularAggregationType
            Aggregation to apply to the inputs.

        """
        from merlin.models.tf.core.combinators import SequentialBlock, TabularBlock

        block = TabularBlock(post=post, aggregation=aggregation) or block

        return SequentialBlock([block, self])

    def repeat_in_parallel(
        self,
        num: int = 1,
        prefix=None,
        names: Optional[List[str]] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional["TabularAggregationType"] = None,
        copies=True,
        residual=False,
        **kwargs,
    ) -> "ParallelBlock":
        """Repeat the block num times in parallel.

        Parameters
        ----------
        num: int
            Number of times to repeat the block.
        prefix: str
            Prefix to use for the names of the blocks.
        names: List[str]
            Names of the blocks.
        post: Block
            Block to use as post-transformation.
        aggregation: TabularAggregationType
            Aggregation to apply to the inputs.
        copies: bool
            Whether to copy the block or not.
        residual: bool
            Whether to use a residual connection or not.

        """
        from merlin.models.tf.core.base import NoOp
        from merlin.models.tf.core.combinators import ParallelBlock

        repeated = {}
        iterator = names if names else range(num)
        if not names and prefix:
            iterator = [f"{prefix}{num}" for num in iterator]
        for name in iterator:
            repeated[str(name)] = self.copy() if copies else self

        if residual:
            repeated["shortcut"] = NoOp()

        return ParallelBlock(repeated, post=post, aggregation=aggregation, **kwargs)

    def connect(
        self,
        *block: Union[tf.keras.layers.Layer, str],
        block_name: Optional[str] = None,
        context: Optional[ModelContext] = None,
    ) -> "SequentialBlock":
        """Connect the block to other blocks sequentially.

        Parameters
        ----------
        block: Union[tf.keras.layers.Layer, str]
            Blocks to connect to.
        block_name: str
            Name of the block.
        context: Optional[ModelContext]
            Context to use for the block.

        """
        from merlin.models.tf.core.combinators import SequentialBlock

        blocks = [self.parse(b) for b in block]

        for b in blocks:
            if isinstance(b, Block):
                if not b.has_schema and self.has_schema:
                    b._schema = self.schema

        output = SequentialBlock(
            [self, *blocks], copy_layers=False, block_name=block_name, context=context
        )

        return output

    def connect_with_residual(
        self,
        block: Union[tf.keras.layers.Layer, str],
        activation=None,
    ) -> "SequentialBlock":
        """Connect the block to other blocks sequentially with a residual connection.

        Parameters
        ----------
        block: Union[tf.keras.layers.Layer, str]
            Blocks to connect to.
        activation: str
            Activation to use for the residual connection.

        """
        from merlin.models.tf.core.combinators import ResidualBlock, SequentialBlock

        _block = self.parse(block)
        residual_block = ResidualBlock(_block, activation=activation)

        if isinstance(self, SequentialBlock):
            self.layers.append(residual_block)

            return self

        return SequentialBlock([self, residual_block], copy_layers=False)

    def connect_with_shortcut(
        self,
        block: Union[tf.keras.layers.Layer, str],
        shortcut_filter: Optional["Filter"] = None,
        post: Optional[BlockType] = None,
        aggregation: Optional["TabularAggregationType"] = None,
        block_outputs_name: Optional[str] = None,
    ) -> "SequentialBlock":
        """Connect the block to other blocks sequentially with a shortcut connection.

        Parameters
        ----------
        block: Union[tf.keras.layers.Layer, str]
            Blocks to connect to.
        shortcut_filter: Filter
            Filter to use for the shortcut connection.
        post: Block
            Block to use as post-transformation.
        aggregation: TabularAggregationType
            Aggregation to apply to the outputs.
        block_outputs_name: str
            Name of the block outputs.
        """
        from merlin.models.tf.core.combinators import SequentialBlock, WithShortcut

        _block = self.parse(block) if not isinstance(block, Block) else block
        residual_block = WithShortcut(
            _block,
            shortcut_filter=shortcut_filter,
            post=post,
            aggregation=aggregation,
            block_outputs_name=block_outputs_name,
            automatic_pruning=False,
        )

        if isinstance(self, SequentialBlock):
            self.layers.append(residual_block)

            return self

        return SequentialBlock([self, residual_block], copy_layers=False)

    def connect_debug_block(self, append=True):
        """Connect the block to a debug block.

        Parameters
        ----------
        append: bool
            Whether to append the debug block to the block or to prepend it.
        """
        from merlin.models.tf.core.base import Debug
        from merlin.models.tf.core.combinators import SequentialBlock

        if not append:
            return SequentialBlock([Debug(), self])

        return self.connect(Debug())

    def connect_branch(
        self,
        *branches: Union["Block", "PredictionTask", str],
        add_rest=False,
        post: Optional[BlockType] = None,
        aggregation: Optional["TabularAggregationType"] = None,
        **kwargs,
    ) -> "SequentialBlock":
        """Connect the block to one or multiple branches.

        Parameters
        ----------
        branches: Union[Block, PredictionTask, str]
            Blocks to connect to.
        add_rest: bool
            Whether to add the rest of the block to the branches.
        post: Block
            Block to use as post-transformation.
        aggregation: TabularAggregationType
            Aggregation to apply to the outputs.

        """
        from merlin.models.tf.core.combinators import Filter, ParallelBlock, SequentialBlock

        _branches = [self.parse(b) for b in branches]

        all_features = []
        for branch in _branches:
            if getattr(branch, "set_schema", None):
                branch.set_schema(self.schema)
            if isinstance(branch, SequentialBlock):
                filter_features = branch.filter_features
                if filter_features:
                    all_features.extend(filter_features)

        if add_rest:
            if not self.schema:
                raise ValueError("Schema is required to add rest features.")
            rest_features = self.schema.without(list(set([str(f) for f in all_features])))
            rest_block = SequentialBlock([Filter(rest_features)])
            _branches.append(rest_block)

        return SequentialBlock(
            [self, ParallelBlock(*_branches, post=post, aggregation=aggregation, **kwargs)]
        )

    def select_by_name(self, name: str) -> Optional["Block"]:
        if name == self.name:
            return self
        return None

    def select_by_tag(self, tags: Tags) -> Optional["Block"]:
        selected_schema = self.schema.select_by_tag(tags)
        if selected_schema:
            return self
        return None

    def copy(self):
        return self.from_config(self.get_config())

    def _maybe_propagate_context(self, input_shapes):
        if getattr(self, "_context", None) and not self.context.built:
            for module in self.submodules:
                if hasattr(module, "_set_context") and not getattr(module, "context", False):
                    module._set_context(self.context)
            self._need_to_call_context = True
            self.context.build(input_shapes)

    def __rrshift__(self, other):
        return right_shift_layer(self, other)

    def get_config(self):
        config = super().get_config()
        return config


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class NoOp(tf.keras.layers.Layer):
    def call(self, inputs, **kwargs):
        return inputs

    def compute_output_shape(self, input_shape):
        return input_shape


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class Debug(tf.keras.layers.Layer):
    def call(self, inputs, **kwargs):
        return inputs

    def compute_output_shape(self, input_shape):
        return input_shape


def name_fn(name, inp):
    return "/".join([name, inp]) if name else None


MetricOrMetricClass = Union[tf.keras.metrics.Metric, Type[tf.keras.metrics.Metric]]
MetricOrMetrics = Union[Sequence[MetricOrMetricClass], MetricOrMetricClass]


@dataclass
class EmbeddingWithMetadata:
    embeddings: tf.Tensor
    metadata: Dict[str, tf.Tensor]


def is_input_block(block: Block) -> bool:
    is_defined = True if block else False
    return is_defined and getattr(block, "is_input", False)


def has_input_block(block: Block) -> bool:
    from merlin.models.tf.core.combinators import SequentialBlock

    if isinstance(block, SequentialBlock):
        return block.inputs is not None and is_input_block(block.inputs)
    return is_input_block(block.inputs)


def _output_metrics(metrics):
    if len(metrics) == 1:
        return metrics[list(metrics.keys())[0]]

    return metrics


def right_shift_layer(self, other):
    from merlin.models.tf.core.combinators import Filter, SequentialBlock

    if isinstance(other, (list, Tags)):
        left_side = [Filter(other)]
    else:
        left_side = other.layers if isinstance(other, SequentialBlock) else [other]
    right_side = self.layers if isinstance(self, SequentialBlock) else [self]

    return SequentialBlock(left_side + right_side)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import abc
from typing import Dict, List, NamedTuple, Optional, Sequence, Union

import tensorflow as tf

from merlin.models.utils.registry import Registry, RegistryMixin

EMBEDDING_KEY = "__embedding__"


class Candidate(NamedTuple):
    """Store candidate id and their metadata

    Parameters
    ----------
    id : tf.Tensor
        The tensor of item ids
    metadata:
        dictionary of tensors containing meta information
        about items such as item embeddings and item category
    """

    id: tf.Tensor
    metadata: Dict[str, tf.Tensor]

    @property
    def embedding(self) -> tf.Tensor:
        return self.metadata[EMBEDDING_KEY]

    @property
    def has_embedding(self) -> bool:
        return EMBEDDING_KEY in self.metadata

    def with_embedding(self, embedding: tf.Tensor) -> "Candidate":
        self.metadata[EMBEDDING_KEY] = embedding

        return self

    def __add__(self, other):
        metadata = {}
        for key in self.metadata:
            if key in other.metadata:
                metadata[key] = _list_to_tensor([self.metadata[key], other.metadata[key]])

        return Candidate(
            id=_list_to_tensor([self.id, other.id]),
            metadata=metadata,
        )

    @property
    def shape(self) -> "Candidate":
        return Candidate(self.id.shape, {key: val.shape for key, val in self.metadata.items()})

    def __repr__(self):
        metadata = {key: str(val) for key, val in self.metadata.items()}

        return f"Candidate({self.id}, {metadata})"

    def __str__(self):
        metadata = {key: str(val) for key, val in self.metadata.items()}

        return f"Candidate({self.id}, {metadata})"

    def __eq__(self, other) -> bool:
        if self.id.shape != other.id.shape:
            return False

        return self.id.ref() == other.id.ref()

    def get_config(self):
        return {
            "ids": self.id.as_list() if self.id else None,
            "metadata": {key: val.as_list() for key, val in self.metadata.items()},
        }

    @classmethod
    def from_config(cls, config):
        ids = tf.TensorShape(config["config"]["id"])
        metadata = {key: tf.TensorShape(val) for key, val in config["config"]["metadata"].items()}

        return cls(ids, metadata)


negative_sampling_registry: Registry = Registry.class_registry("tf.negative_sampling")


class CandidateSampler(tf.keras.layers.Layer, RegistryMixin["CandidateSampler"], abc.ABC):
    """Base-class for negative sampling

    Parameters
    ----------
    max_num_samples : int
        The number of maximum samples to store

    Returns
    -------
    Items
        The sampled ids and their metadata
    """

    registry = negative_sampling_registry

    def __init__(
        self,
        max_num_samples: Optional[int] = None,
        **kwargs,
    ):
        super(CandidateSampler, self).__init__(**kwargs)
        self.set_max_num_samples(max_num_samples)

    def call(
        self, items: Candidate, features=None, targets=None, training=False, testing=False
    ) -> Candidate:
        if training:
            self.add(items)
        items = self.sample()

        return items

    @abc.abstractmethod
    def add(self, items: Candidate):
        raise NotImplementedError()

    @abc.abstractmethod
    def sample(self) -> Candidate:
        raise NotImplementedError()

    @property
    def max_num_samples(self) -> int:
        return self._max_num_samples

    def set_max_num_samples(self, value) -> None:
        self._max_num_samples = value


def _list_to_tensor(input_list: List[tf.Tensor]) -> tf.Tensor:
    output: tf.Tensor

    if len(input_list) == 1:
        output = input_list[0]
    else:
        output = tf.concat(input_list, axis=0)

    return output


ItemSamplersType = Union[CandidateSampler, Sequence[Union[CandidateSampler, str]], str]


def parse_negative_samplers(negative_sampling: ItemSamplersType):
    """
    Parse the negative sampling strategies and returns
    the corresponding list of samplers.
    """
    if negative_sampling is None:
        negative_sampling = []
    if not isinstance(negative_sampling, (list, tuple)):
        negative_sampling = [negative_sampling]
    negative_sampling = [CandidateSampler.parse(s) for s in list(negative_sampling)]

    return negative_sampling



import pytest
from testbook import testbook

from merlin.core.dispatch import HAS_GPU
from tests.conftest import REPO_ROOT

pytest.importorskip("xgboost")
pytest.importorskip("lightfm")
pytest.importorskip("implicit")


@testbook(
    REPO_ROOT / "examples/07-Train-traditional-ML-models-using-the-Merlin-Models-API.ipynb",
    execute=False,
)
@pytest.mark.skipif(not HAS_GPU, reason="No GPU available")
def test_func(tb):
    tb.execute()
    xgboost_metrics = tb.ref("metrics")
    implicit_metrics = tb.ref("implicit_metrics")

    assert xgboost_metrics.keys() == {"logloss"}
    assert sorted(implicit_metrics) == ["auc@10", "map@10", "ndcg@10", "precision@10"]



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Optional

import tensorflow as tf

from merlin.models.tf.blocks.mlp import MLPBlock
from merlin.models.tf.core.aggregation import StackFeatures
from merlin.models.tf.core.base import Block
from merlin.models.tf.core.combinators import MapValues, ParallelBlock, SequentialBlock
from merlin.models.tf.core.tabular import Filter
from merlin.models.tf.inputs.base import InputBlockV2
from merlin.models.tf.inputs.embedding import Embeddings
from merlin.models.tf.transforms.features import CategoryEncoding, ToSparse
from merlin.schema import Schema, Tags

_INTERACTION_TYPES = (None, "field_all", "field_each", "field_interaction")


@tf.keras.utils.register_keras_serializable(package="merlin_models")
class DotProductInteraction(tf.keras.layers.Layer):
    """
    Layer implementing the factorization machine style feature
    interaction layer suggested by the DLRM and DeepFM architectures,
    generalized to include a dot-product version of the parameterized
    interaction suggested by the FiBiNet architecture (which normally
    uses element-wise multiplication instead of dot product). Maps from
    tensors of shape `(batch_size, num_features, embedding_dim)` to
    tensors of shape `(batch_size, (num_features - 1)*num_features // 2)`
    if `self_interaction` is `False`, otherwise `(batch_size, num_features**2)`.

    Parameters
    ------------------------
    interaction_type: {}
        The type of feature interaction to use. `None` defaults to the
        standard factorization machine style interaction, and the
        alternatives use the implementation defined in the FiBiNet
        architecture (with the element-wise multiplication replaced
        with a dot product).
    self_interaction: bool
        Whether to calculate the interaction of a feature with itself.
    """.format(
        _INTERACTION_TYPES
    )

    def __init__(self, interaction_type=None, self_interaction=False, name=None, **kwargs):
        if interaction_type not in _INTERACTION_TYPES:
            raise ValueError("Unknown interaction type {}".format(interaction_type))
        self.interaction_type = interaction_type
        self.self_interaction = self_interaction
        super(DotProductInteraction, self).__init__(name=name, **kwargs)

    def build(self, input_shape):
        if self.interaction_type is None:
            self.built = True
            return

        kernel_shape = [input_shape[2], input_shape[2]]
        if self.interaction_type in _INTERACTION_TYPES[2:]:
            idx = _INTERACTION_TYPES.index(self.interaction_type)
            for _ in range(idx - 1):
                kernel_shape.insert(0, input_shape[1])

        self.kernel = self.add_weight(
            name="bilinear_interaction_kernel",
            shape=kernel_shape,
            initializer="glorot_normal",
            trainable=True,
        )
        self.built = True

    def call(self, inputs):
        right = inputs

        # first transform v_i depending on the interaction type
        if self.interaction_type is None:
            left = inputs
        elif self.interaction_type == "field_all":
            left = tf.matmul(inputs, self.kernel)
        elif self.interaction_type == "field_each":
            left = tf.einsum("b...k,...jk->b...j", inputs, self.kernel)
        else:
            left = tf.einsum("b...k,f...jk->bf...j", inputs, self.kernel)

        # do the interaction between v_i and v_j
        # output shape will be (batch_size, num_features, num_features)
        if self.interaction_type != "field_interaction":
            interactions = tf.matmul(left, right, transpose_b=True)
        else:
            interactions = tf.einsum("b...jk,b...k->b...j", left, right)

        # mask out the appropriate area
        ones = tf.reduce_sum(tf.zeros_like(interactions), axis=0) + 1
        mask = tf.linalg.band_part(ones, 0, -1)  # set lower diagonal to zero
        if not self.self_interaction:
            mask = mask - tf.linalg.band_part(ones, 0, 0)  # get rid of diagonal
        mask = tf.cast(mask, tf.bool)
        x = tf.boolean_mask(interactions, mask, axis=1)

        # masking destroys shape information, set explicitly
        x.set_shape(self.compute_output_shape(inputs.shape))
        return x

    def compute_output_shape(self, input_shape):
        if self.self_interaction:
            output_dim = input_shape[1] ** 2
        else:
            output_dim = input_shape[1] * (input_shape[1] - 1) // 2

        return input_shape[0], output_dim

    def get_config(self):
        return {
            "interaction_type": self.interaction_type,
            "self_interaction": self.self_interaction,
        }


class XDeepFmOuterProduct(tf.keras.layers.Layer):
    """
    Layer implementing the outer product transformation used in
    the Compressed Interaction Network (CIN) proposed in
    in https://arxiv.org/abs/1803.05170. Treats the feature dimension
    H_k of a B x H_k x D feature embedding tensor as a feature map
    of the D embedding elements, and computes element-wise multiplication
    interaction between these maps and those from an initial input tensor
    x_0 before taking the inner product with a parameter matrix.

    Parameters
    ------------
    dim : int
      Feature dimension of the layer. Output will be of shape
      (batch_size, dim, embedding_dim)
    """

    def __init__(self, dim, **kwargs):
        self.dim = dim
        super().__init__(**kwargs)

    def build(self, input_shapes):
        if not isinstance(input_shapes[0], (tuple, tf.TensorShape)):
            raise ValueError("Should be called on a list of inputs.")
        if len(input_shapes) != 2:
            raise ValueError("Should only have two inputs, found {}".format(len(input_shapes)))
        for shape in input_shapes:
            if len(shape) != 3:
                raise ValueError("Found shape {} without 3 dimensions".format(shape))
        if input_shapes[0][-1] != input_shapes[1][-1]:
            raise ValueError(
                "Last dimension should match, found dimensions {} and {}".format(
                    input_shapes[0][-1], input_shapes[1][-1]
                )
            )

        # H_k x H_{k-1} x m
        shape = (self.dim, input_shapes[0][1], input_shapes[1][1])
        self.kernel = self.add_weight(
            name="kernel", initializer="glorot_uniform", trainable=True, shape=shape
        )
        self.built = True

    def call(self, inputs):
        """
        Parameters
        ------------
        inputs : array-like(tf.Tensor)
          The two input tensors, the first of which should be the
          output of the previous layer, and the second of which
          should be the input to the CIN.
        """
        x_k_minus_1, x_0 = inputs

        # need to do shape manipulations so that we
        # can do element-wise multiply
        x_k_minus_1 = tf.expand_dims(x_k_minus_1, axis=2)  # B x H_{k-1} x 1 x D
        x_k_minus_1 = tf.tile(x_k_minus_1, [1, 1, x_0.shape[1], 1])  # B x H_{k-1} x m x D
        x_k_minus_1 = tf.transpose(x_k_minus_1, (1, 0, 2, 3))  # H_{k-1} x B x m x D
        z_k = x_k_minus_1 * x_0  # H_{k-1} x B x m x D
        z_k = tf.transpose(z_k, (1, 0, 2, 3))  # B x H_{k-1} x m x D

        # now we need to map to B x H_k x D
        x_k = tf.tensordot(self.kernel, z_k, axes=[[1, 2], [1, 2]])
        x_k = tf.transpose(x_k, (1, 0, 2))
        return x_k

    def compute_output_shape(self, input_shapes):
        return (input_shapes[0][0], self.dim, input_shapes[0][2])


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class FMPairwiseInteraction(Block):
    """Compute pairwise (2nd-order) feature interactions like defined in
    Factorized Machine [1].

    References
    ----------
    [1] Steffen, Rendle, "Factorization Machines" IEEE International
    Conference on Data Mining, 2010. https://ieeexplore.ieee.org/document/5694074
    """

    def call(self, inputs: tf.Tensor, **kwargs) -> tf.Tensor:
        """
        Parameters
        ----------
        inputs : array-like(tf.Tensor)
          A 3-D tensor of shape (bs, n_features, embedding_dim)
          containing the stacked embeddings of input features.

        Returns
        -------
        A 2-D tensor of shape (bs, K) containing pairwise interactions
        """
        assert len(inputs.shape) == 3, "inputs should be a 3-D tensor"

        # sum_square part
        summed_square = tf.square(tf.reduce_sum(inputs, 1))

        # square_sum part
        squared_sum = tf.reduce_sum(tf.square(inputs), 1)

        # second order
        return 0.5 * tf.subtract(summed_square, squared_sum)

    def compute_output_shape(self, input_shapes):
        if len(input_shapes) != 3:
            raise ValueError("Found shape {} without 3 dimensions".format(input_shapes))
        return (input_shapes[0], input_shapes[2])


def FMBlock(
    schema: Schema,
    fm_input_block: Optional[Block] = None,
    wide_input_block: Optional[Block] = None,
    wide_logit_block: Optional[Block] = None,
    factors_dim: Optional[int] = None,
    **kwargs,
) -> tf.Tensor:
    """Implements the Factorization Machine, as introduced in [1].
    It consists in the sum of a wide component that weights each
    feature individually and a 2nd-level feature interaction using
    factors.

    References
    ----------
    [1] Steffen, Rendle, "Factorization Machines" IEEE International
    Conference on Data Mining, 2010. https://ieeexplore.ieee.org/document/5694074

    Parameters
    ----------
    schema : Schema
        The schema of input features
    fm_input_block : Optional[Block], by default None
        The input block for the 2nd-order feature interaction in Factorization Machine.
        Only categorical features will be used by this block, as it computes
        dot product between all paired combinations of embedding values.
        If not provided, an InputBlockV2 is instantiated based on schema.
        Note: All features (including continuous) are considered in the
        1st-order (wide) part which uses another input block.
    wide_input_block: Optional[Block], by default None
        The input for the wide block. If not provided,
        creates a default block that encodes categorical features
        with one-hot / multi-hot representation and also includes the continuous features.
    wide_logit_block: Optional[Block], by default None
        The output layer of the wide input. The last dimension needs to be 1.
        You might want to provide your own output logit block if you want to add
        dropout or kernel regularization to the wide block.
    factors_dim : Optional[int], optional
        If fm_input_block is not provided, the factors_dim is used to define the
        embeddings dim to instantiate InputBlockV2, by default None
    Returns
    -------
    tf.Tensor
        Returns a 2D tensor (batch size, 1) with the sum of the wide component
        and 2nd-order interaction component of FM
    """

    cat_schema = schema.select_by_tag(Tags.CATEGORICAL)
    cont_schema = schema.select_by_tag(Tags.CONTINUOUS)

    wide_input_block = wide_input_block or ParallelBlock(
        {
            "categorical": CategoryEncoding(cat_schema, output_mode="multi_hot", sparse=True),
            "continuous": Filter(cont_schema).connect(ToSparse()),
        },
        aggregation="concat",
    )

    wide_logit_block = wide_logit_block or MLPBlock([1], activation="linear", use_bias=True)
    first_order = wide_input_block.connect(wide_logit_block)

    fm_input_block = fm_input_block or InputBlockV2(
        cat_schema,
        categorical=Embeddings(cat_schema, dim=factors_dim),
        aggregation=None,
    )
    pairwise_interaction = SequentialBlock(
        Filter(cat_schema),
        fm_input_block,
        FMPairwiseInteraction().prepare(aggregation=StackFeatures(axis=-1)),
        MapValues(tf.keras.layers.Lambda(lambda x: tf.reduce_sum(x, axis=1, keepdims=True))),
    )

    fm_block = ParallelBlock([first_order, pairwise_interaction], aggregation="element-wise-sum")

    return fm_block



from typing import Optional, Union

import tensorflow as tf
from tensorflow.keras.layers import Layer

import merlin.io
from merlin.core.dispatch import DataFrameType
from merlin.models.tf.core.base import Block, block_registry
from merlin.models.tf.core.prediction import Prediction, TopKPrediction
from merlin.models.tf.outputs.base import MetricsFn, ModelOutput
from merlin.models.tf.outputs.classification import default_categorical_prediction_metrics
from merlin.models.tf.utils import tf_utils
from merlin.schema import Schema


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class TopKLayer(Layer):
    """Base class for Top-K index.
    The objective is to retrieve top-k scores and candidates ids
    for a given query.

    Parameters:
    -----------
    k: int
        Number of candidates to return

    """

    def __init__(self, k: int, **kwargs) -> None:
        """Initializes the base class."""
        super().__init__(**kwargs)
        self._k = k

    def index(self, candidates: tf.Tensor, identifiers: Optional[tf.Tensor] = None) -> "TopKLayer":
        """Set the index used for retrieving the top-k candidates.
        When called multiple times the existing index will be dropped and a new one
        created.
        Parameters:
        -----------
            candidates: tensor of candidate embeddings.
            identifiers: Optional tensor of candidate identifiers. If
                given, these will be used as identifiers of top candidates returned
                when performing searches. If not given, indices into the candidates
                tensor will be returned instead.

        """
        raise NotImplementedError()

    def index_from_dataset(
        self, data: merlin.io.Dataset, check_unique_ids: bool = True
    ) -> "TopKLayer":
        """Builds the top-k retrieval index from a merlin dataset.
        Parameters
        ----------
        data : merlin.io.Dataset
            The dataset with the pre-trained item embeddings
        check_unique_ids : bool, optional
            Whether to check if `data` has unique indices, by default True
        Returns
        -------
        TopKLayer
            return the class with retrieval index
        """
        ids, values = self.extract_ids_embeddings(data, check_unique_ids)
        return self.index(candidates=values, identifiers=ids)

    @staticmethod
    def _check_unique_ids(data: DataFrameType):
        if data.index.to_series().nunique() != data.shape[0]:
            raise ValueError("Please make sure that `data` contains unique indices")

    def extract_ids_embeddings(self, data: merlin.io.Dataset, check_unique_ids: bool = True):
        """Extract tensors of candidates and indices from the merlin dataset
        Parameters
        ----------
        data : merlin.io.Dataset
            The dataset with the pre-trained candidates embeddings,
            indexed by the candidates identifiers.
        check_unique_ids : bool, optional
            Whether to check if `data` has unique indices, by default True
        """
        if hasattr(data, "to_ddf"):
            data = data.to_ddf()
        if check_unique_ids:
            self._check_unique_ids(data=data)
        values = tf_utils.df_to_tensor(data)
        ids = tf_utils.df_to_tensor(data.index)

        if len(ids.shape) == 2:
            ids = tf.squeeze(ids)
        return ids, values

    def call(self, inputs: tf.Tensor, targets=None, testing=False, **kwargs) -> tf.Tensor:
        """Method to return the tuple of top-k (ids, scores)"""
        raise NotImplementedError()

    def _score(self, queries: tf.Tensor, candidates: tf.Tensor) -> tf.Tensor:
        """Computes the standard dot product score from queries and candidates."""
        return tf.matmul(queries, candidates, transpose_b=True)

    @classmethod
    def from_config(cls, config, custom_objects=None):
        return super().from_config(config)

    def get_config(self):
        config = super().get_config()
        config["k"] = self._k
        return config


@Block.registry.register("brute-force-topk")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class BruteForce(TopKLayer):
    """Top-k layer that performs a brute force search over the candidates index."""

    def __init__(self, k: int = 10, name: Optional[str] = None, **kwargs):
        """Initializes the layer.

        Parameters:
        -----------
        k: int
            Default threshold for top-k retrieval. By default 10
        name: Optional[str]
            Name of the layer. By default None
        """

        super().__init__(k=k, name=name, **kwargs)
        self._candidates = kwargs.pop("_candidates", None)

    def index(self, candidates: tf.Tensor, identifiers: Optional[tf.Tensor] = None) -> "BruteForce":

        tf.assert_equal(
            tf.rank(candidates), 2, f"candidates must be 2-D tensor (got {candidates.shape})"
        )

        if identifiers is None:
            identifiers = tf.range(candidates.shape[0])

        tf.assert_equal(
            tf.shape(candidates)[0],
            tf.shape(identifiers)[0],
            "The candidates and identifiers tensors must have the same number of rows "
            f"(got {candidates.shape[0]} candidates rows and {identifiers.shape[0]} "
            "identifier rows).",
        )

        self._ids = self.add_weight(
            name="ids",
            dtype=tf.int32,
            shape=identifiers.shape,
            initializer=tf.keras.initializers.Constant(value=tf.cast(identifiers, tf.int32)),
            trainable=False,
        )

        self._candidates = self.add_weight(
            name="candidates",
            dtype=tf.float32,
            shape=candidates.shape,
            initializer=tf.keras.initializers.Zeros(),
            trainable=False,
        )

        self._ids.assign(tf.cast(identifiers, tf.int32))
        self._candidates.assign(tf.cast(candidates, tf.float32))
        return self

    def call(
        self,
        inputs,
        targets=None,
        testing=False,
    ) -> Union[Prediction, TopKPrediction]:
        """Compute the scores between the query inputs and all indexed candidates,
        then retrieve the top-k candidates with the highest scores.

        Parameters
        ----------
        inputs : tf.Tensor
            The query embeddings representation
        targets: tf.Tensor
            The tensor of positive candidates
        testing: bool
            Flag that indicates whether in evaluation mode, by default False
        """
        if self._candidates is None:
            raise ValueError(
                "You should call the `index` method first to " "set the _candidates index."
            )

        tf.assert_equal(
            tf.shape(inputs)[1],
            tf.shape(self._candidates)[1],
            "Query and candidates vectors must have the same embedding size "
            f"(got query dimension of {tf.shape(inputs)[1]} and candidates"
            f" dimension of {tf.shape(self._candidates)[1]} ",
        )
        scores = self._score(inputs, self._candidates)
        top_scores, top_ids = tf.math.top_k(scores, k=self._k)
        if testing:
            assert targets is not None, ValueError(
                "Targets should be provided during the evaluation mode"
            )
            targets = tf.cast(tf.squeeze(targets), tf.int32)
            targets = tf.cast(tf.expand_dims(targets, -1) == top_ids, tf.float32)
            targets = tf.reshape(targets, tf.shape(top_scores))
            return Prediction(top_scores, targets)
        return TopKPrediction(top_scores, top_ids)

    def compute_output_shape(self, input_shape):
        return (
            tf.TensorShape((input_shape[0], self._k)),
            tf.TensorShape((input_shape[0], self._k)),
        )


@tf.keras.utils.register_keras_serializable(package="merlin.models")
class TopKOutput(ModelOutput):
    """Prediction block for top-k evaluation
    Parameters
    ----------
    to_call:  Union[str, TopKLayer]
       The top-k layer to use for retrieving top-k candidates ids and scores
    item_dataset:  merlin.io.Dataset,
        Dataset of the pretrained candidates embeddings to use for the top-k index.
    k: int, optional
        Number of candidates to return, by default 10
    target: Union[str, Schema], optional
        The name of the target. If a Schema is provided, the target is inferred from the schema.
    pre: Optional[Block], optional
        Optional block to transform predictions before computing the binary logits,
        by default None
    post: Optional[Block], optional
        Optional block to transform the binary logits,
        by default None
    name: str, optional
        The name of the task.
    logits_temperature: float, optional
        Parameter used to reduce model overconfidence, so that logits / T.
        by default 1.
    default_loss: Union[str, tf.keras.losses.Loss], optional
        Default loss to use for binary-classification
        by 'categorical_crossentropy'
    default_metrics_fn: Callable
        A function returning the list of default metrics
        to use for ranking evaluation
    """

    def __init__(
        self,
        to_call: Union[str, TopKLayer],
        candidates: Union[tf.Tensor, merlin.io.Dataset] = None,
        k: int = 10,
        target: Optional[Union[str, Schema]] = None,
        pre: Optional[Layer] = None,
        post: Optional[Layer] = None,
        logits_temperature: float = 1.0,
        name: Optional[str] = None,
        default_loss: Union[str, tf.keras.losses.Loss] = "categorical_crossentropy",
        default_metrics_fn: MetricsFn = default_categorical_prediction_metrics,
        **kwargs,
    ):

        if isinstance(to_call, str):
            if candidates is None:
                raise ValueError(
                    "You should provide the dataset of pre-trained embeddings"
                    " when `to_call` is the top-k strategy name "
                )
            if isinstance(candidates, merlin.io.Dataset):
                to_call = block_registry.parse(to_call).index_from_dataset(candidates)
            else:
                to_call = block_registry.parse(to_call).index(candidates)
            to_call._k = k

        assert isinstance(to_call, TopKLayer), ValueError(
            "TopKOutput requires `to_call` to be an instance of a `TopKLayer`,"
            f" {type(to_call)} is provided"
        )
        self.to_call = to_call
        super().__init__(
            to_call=to_call,
            default_loss=default_loss,
            default_metrics_fn=default_metrics_fn,
            name=name,
            target=target,
            pre=pre,
            post=post,
            logits_temperature=logits_temperature,
            **kwargs,
        )



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pytest
import tensorflow as tf
from tensorflow.keras import regularizers

import merlin.models.tf as ml
from merlin.io import Dataset


@pytest.mark.parametrize("dim", [32, 64])
@pytest.mark.parametrize("activation", ["relu", "tanh"])
@pytest.mark.parametrize("dropout", [None, 0.5])
@pytest.mark.parametrize(
    "normalization", [None, "batch_norm", tf.keras.layers.BatchNormalization()]
)
def test_mlp_block_yoochoose(
    testing_data: Dataset,
    dim,
    activation,
    dropout,
    normalization,
    kernel_initializer="glorot_uniform",
    bias_initializer="zeros",
    kernel_regularizer="l2",
    bias_regularizer="l1",
    activity_regularizer=regularizers.l2(1e-4),
):
    inputs = ml.InputBlock(testing_data.schema)

    mlp = ml.MLPBlock(
        [dim],
        activation=activation,
        dropout=dropout,
        normalization=normalization,
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        bias_regularizer=bias_regularizer,
        activity_regularizer=activity_regularizer,
    )
    body = ml.SequentialBlock([inputs, mlp])

    outputs = body(ml.sample_batch(testing_data, batch_size=100, include_targets=False))

    assert list(outputs.shape) == [100, dim]
    assert mlp.layers[0].units == dim
    assert mlp.layers[0].dense.activation.__name__ == activation

    if dropout:
        assert mlp.layers[1].rate == dropout
    if normalization:
        if normalization == "batch_norm":
            normalization = tf.keras.layers.BatchNormalization()

        assert mlp.layers[-1].__class__.__name__ == normalization.__class__.__name__


@pytest.mark.parametrize("no_activation_last_layer", [False, True])
@pytest.mark.parametrize("dims", [[32], [64, 32]])
def test_mlp_block_no_activation_last_layer(no_activation_last_layer, dims):
    mlp = ml.MLPBlock(dims, activation="relu", no_activation_last_layer=no_activation_last_layer)

    for idx, layer in enumerate(mlp.layers):
        if no_activation_last_layer and idx == len(mlp.layers) - 1:
            assert layer.dense.activation.__name__ == "linear"
        else:
            assert layer.dense.activation.__name__ == "relu"


@pytest.mark.parametrize("run_eagerly", [True])
def test_mlp_model_save(ecommerce_data: Dataset, run_eagerly: bool, tmp_path):
    model = ml.Model.from_block(
        ml.MLPBlock(
            [64], kernel_regularizer=regularizers.l2(1e-1), bias_regularizer=regularizers.l2(1e-1)
        ),
        ecommerce_data.schema,
    )

    model.compile(optimizer="adam", run_eagerly=run_eagerly)
    model.fit(ecommerce_data, batch_size=50, epochs=1)
    model.save(str(tmp_path))

    copy_model = tf.keras.models.load_model(str(tmp_path))
    copy_model.compile(optimizer="adam", run_eagerly=run_eagerly)

    assert isinstance(copy_model, tf.keras.Model)


@pytest.mark.parametrize(
    "arguments", [([32], "relu"), ([64, 32], "relu"), ([32, 16], ["relu", "linear"])]
)
def test_mlp_block_dense_layer_activation(arguments):
    mlp = ml.MLPBlock(dimensions=arguments[0], activation=arguments[1])

    for idx, layer in enumerate(mlp.layers):
        activation_idx = arguments[1] if isinstance(arguments[1], str) else arguments[1][idx]
        assert layer.dense.activation.__name__ == activation_idx


def test_mlp_block_activation_dimensions_length_mismatch():
    with pytest.raises(ValueError) as excinfo:
        _ = ml.MLPBlock(dimensions=[32], activation=["relu", "linear"])
    assert "Activation and Dimensions length mismatch." in str(excinfo.value)



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from dataclasses import dataclass
from typing import Any, Dict, Sequence, Union

from merlin.schema import ColumnSchema, Schema, Tags


class SchemaMixin:
    REQUIRES_SCHEMA = False

    def set_schema(self, schema=None):
        self.check_schema(schema=schema)

        if schema and not getattr(self, "_schema", None):
            self._schema = schema

        return self

    @property
    def schema(self) -> Schema:
        if not self.has_schema:
            raise ValueError(f"{self.__class__.__name__} does not have a schema.")

        return self._schema

    @property
    def has_schema(self):
        return getattr(self, "_schema", None) is not None

    @schema.setter  # type: ignore
    def schema(self, value):
        if value:
            self.set_schema(value)
        else:
            self._schema = value

    def check_schema(self, schema=None):
        if self.REQUIRES_SCHEMA and not getattr(self, "_schema", None) and not schema:
            raise ValueError(f"{self.__class__.__name__} requires a schema.")

    def __call__(self, *args, **kwargs):
        self.check_schema()

        return super().__call__(*args, **kwargs)

    def _maybe_set_schema(self, input, schema):
        if input and getattr(input, "set_schema", None):
            input.set_schema(schema)

    def get_item_ids_from_inputs(self, inputs):
        return inputs[self.schema.select_by_tag(Tags.ITEM_ID).first.name]

    def get_padding_mask_from_item_id(self, inputs, pad_token=0):
        item_id_inputs = self.get_item_ids_from_inputs(inputs)
        if len(item_id_inputs.shape) != 2:
            raise ValueError(
                "To extract the padding mask from item id tensor "
                "it is expected to have 2 dims, but it has {} dims.".format(item_id_inputs.shape)
            )
        return self.get_item_ids_from_inputs(inputs) != pad_token


def requires_schema(module):
    module.REQUIRES_SCHEMA = True

    return module


@dataclass(frozen=True)
class Feature:
    """
    A feature containing its schema and data.
    """

    schema: ColumnSchema
    value: Any


class FeatureCollection:
    """
    A collection of features containing their schemas and data.
    """

    def __init__(self, schema: Schema, values: Dict[str, Any]):
        self.values = values
        self.schema = schema

    def with_schema(self, schema: Schema) -> "FeatureCollection":
        """
        Create a new FeatureCollection with the same data and an updated Schema.

        Parameters
        ----------
        schema : Schema
            Schema to be applied to FeatureCollection

        Returns
        -------
        FeatureCollection
            New collection of features with updated Schema
        """
        return FeatureCollection(schema, self.values)

    def select_by_name(self, names: Union[str, Sequence[str]]) -> "FeatureCollection":
        """
        Create a new FeatureCollection with only the features that match the provided names.

        Parameters
        ----------
        names : string, [string]
            Names of the features to select.

        Returns
        -------
        FeatureCollection
            A collection of the features that match the provided names
        """
        sub_schema = self.schema.select_by_name(names)
        sub_values = {name: self.values[name] for name in sub_schema.column_names}

        return FeatureCollection(sub_schema, sub_values)

    def select_by_tag(
        self, tags: Union[str, Tags, Sequence[str], Sequence[Tags]]
    ) -> "FeatureCollection":
        """
        Create a new FeatureCollection with only the features that match the provided tags.

        Parameters
        ----------
        tags: Union[str, Tags, Sequence[str], Sequence[Tags]]
            Tags or tag strings of the features to select
        Returns
        -------
        FeatureCollection
            A collection of the features that match the provided tags
        """
        sub_schema = self.schema.select_by_tag(tags)
        sub_values = {name: self.values[name] for name in sub_schema.column_names}

        return FeatureCollection(sub_schema, sub_values)

    def __getitem__(self, feature_name: str) -> Feature:
        return Feature(self.schema.column_schemas[feature_name], self.values[feature_name])



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


def is_nvtabular_available() -> bool:
    try:
        import nvtabular
    except ImportError:
        nvtabular = None
    return nvtabular is not None


def is_gpu_dataloader_available() -> bool:
    try:
        import cudf
        import cupy
    except ImportError:
        cudf = None
        cupy = None
    return cudf is not None and cupy is not None


def is_pyarrow_available() -> bool:
    try:
        import pyarrow
    except ImportError:
        pyarrow = None
    return pyarrow is not None


def is_transformers_available() -> bool:
    try:
        import transformers
    except ImportError:
        transformers = None
    return transformers is not None



#
# Copyright (c) 2021, NVIDIA CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Optional

import tensorflow as tf

from merlin.models.tf.outputs.sampling.base import Candidate, CandidateSampler


@CandidateSampler.registry.register("in-batch")
@tf.keras.utils.register_keras_serializable(package="merlin.models")
class InBatchSamplerV2(CandidateSampler):
    """Provides in-batch sampling [1]_ for two-tower item retrieval
    models. The implementation is very simple, as it
    just returns the current item embeddings and metadata, but it is necessary to have
    `InBatchSampler` under the same interface of other more advanced samplers
    (e.g. `CachedCrossBatchSampler`).
    In a nutshell, for a given (user,item) embeddings pair, the other in-batch item
    embeddings are used as negative items, rather than computing different embeddings
    exclusively for negative items.
    This is a popularity-biased sampling     as popular items are observed more often
    in training batches.
    P.s. Ignoring the false negatives (negative items equal to the positive ones) is
    managed by `ItemRetrievalScorer(..., sampling_downscore_false_negatives=True)`
    References
    ----------
    .. [1] Yi, Xinyang, et al. "Sampling-bias-corrected neural modeling for large corpus item
       recommendations." Proceedings of the 13th ACM Conference on Recommender Systems. 2019.
    Parameters
    ----------
    batch_size : int, optional
        The batch size. If not set it is inferred when the layer is built (first call())
    """

    def __init__(self, batch_size: Optional[int] = None, **kwargs):
        super().__init__(max_num_samples=batch_size, **kwargs)
        self._last_batch: Optional[Candidate] = None  # type: ignore
        self.set_batch_size(batch_size)

    @property
    def batch_size(self) -> int:
        return self._batch_size

    def set_batch_size(self, value):
        self._batch_size = value
        if value is not None:
            self.set_max_num_samples(value)

    def build(self, items: Candidate) -> None:
        if isinstance(items, dict):
            items = Candidate.from_config(items)
        if self._batch_size is None:
            if isinstance(items, Candidate):
                self.set_batch_size(items.id[0])
            else:
                self.set_batch_size(items[0])

    def add(self, items: Candidate):
        self._last_batch = items

    def call(
        self, items: Candidate, features=None, targets=None, training=False, testing=False
    ) -> Candidate:
        """Returns the item embeddings and item ids from
        the current batch.

        Parameters
        ----------
        items : Items
            The items ids and their embeddings from the current batch
        features : optional
            The metadata with raw input features, by default None
        targets : _type_, optional
            The tensor of targets, by default None
        training : bool, optional
            Flag indicating if on training mode, by default False
        testing : bool, optional
             Flag indicating if on evaluation mode, by default False

        Returns
        -------
        Items
            NamedTuple with the sampled item ids and item metadata
        """
        self.add(items)
        items = self.sample()

        return items

    def sample(self) -> Candidate:
        return self._last_batch

    def get_config(self):
        config = super().get_config()
        config["batch_size"] = self._batch_size

        # TODO: This is a side-effect, could this lead to problems?
        self._last_batch = None

        return config
